"""Module for fetching data from the Outlook API."""

from datetime import date, datetime, timedelta, timezone
from typing import Any

from O365 import Account
from O365.mailbox import MailBox
import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class OutlookCredentials(BaseModel):
    """Checking for values in Outlook credentials dictionary.

    Two key values are held in the Outlook connector:
        - client_id:
        - client_secret:
        - tenant_id:

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    client_id: str
    client_secret: str
    tenant_id: str


class Outlook(Source):
    """Class implementing the Outlook API.

    Documentation for this API is available at:
        https://o365.github.io/python-o365/latest/getting_started.html.
    """

    UTC = timezone.utc

    def __init__(
        self,
        *args: list[Any],
        credentials: dict[str, Any] | None = None,
        config_key: str = "outlook",
        **kwargs: dict[str, Any],
    ):
        """Outlook connector build for fetching Outlook API source.

        Data are fetched from start to end date range. If start or end date are not
        provided then flow fetched data from yesterday by default.

        Args:
            credentials (Optional[OutlookCredentials], optional): Outlook credentials.
                Defaults to None
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "outlook".

        Examples:
            outlook = Outlook(
                config_key=config_key,
            )
            outlook.api_connection(
                mailbox_name=mailbox_name,
                request_retries=request_retries,
                start_date=start_date,
                end_date=end_date,
                limit=limit,
                address_limit=address_limit,
                outbox_list=outbox_list,
            )
            data_frame = outlook.to_df()

        Raises:
            CredentialError: If credentials are not provided in local_config or
                directly as a parameter.
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(OutlookCredentials(**raw_creds))

        super().__init__(*args, credentials=validated_creds, **kwargs)

    @staticmethod
    def _get_subfolders(
        folder_structure: dict,
        folder: MailBox,
        key_concat: str = "",
    ) -> dict[str, list] | None:
        """Retrieve all the subfolder in a MailBox folder.

        Args:
            folder_structure (dict): Dictionary where to save the data.
            folder (MailBox): The MailBox folder from where to extract the subfolders.
            key_concat (str, optional) Previous Mailbox folder structure to add to
                the actual subfolder. Defaults to "".

        Returns:
            Dict[str, List]: `folder_structure` dictionary is returned once
                it is updated.
        """
        if key_concat:
            tmp_key = key_concat.split("|")
            key_concat = key_concat.replace(f"|{tmp_key[-1]}", "")

        for subfolder in folder.get_folders():
            if subfolder:
                folder_structure.update(
                    {
                        "|".join([key_concat, folder.name, subfolder.name]).lstrip(
                            "|"
                        ): subfolder
                    }
                )

        if folder_structure:
            return folder_structure

        return None

    def _get_all_folders(
        self,
        mailbox: MailBox,
    ) -> dict:
        """To retrieve all folders from a Mailbox object.

        Args:
            mailbox (MailBox): Outlook Mailbox object from where to extract all
                folder structure.

        Returns:
            dict: Every single folder and subfolder is returned as
                "parent (sub)folder|(sub)folder": Mailbox.
        """
        dict_folders = self._get_subfolders({}, mailbox)
        final_dict_folders = dict_folders.copy()

        # Get all subfolders.
        while_dict_folders = {"key": "value"}
        while len(while_dict_folders) != 0:
            while_dict_folders = {}
            for key, value in list(dict_folders.items()):
                tmp_dict_folders = self._get_subfolders({}, value, key_concat=key)
                if tmp_dict_folders:
                    final_dict_folders.update(tmp_dict_folders)
                    while_dict_folders.update(tmp_dict_folders)

            dict_folders = while_dict_folders.copy()

        return final_dict_folders

    # TODO: should be refactored.
    def _get_messages_from_mailbox(  # noqa: C901, PLR0912
        self,
        mailbox_name: str,
        dict_folder: dict,
        date_range_start_time: datetime,
        date_range_end_time: datetime,
        limit: int = 10000,
        address_limit: int = 8000,
        outbox_list: list[str] | None = None,
    ) -> list:
        """To retrieve all messages from all the mailboxes passed in the dictionary.

        Args:
            mailbox_name (str): Mailbox name.
            dict_folder (dict): Mailboxes dictionary holder, with the following
                structure: "parent (sub)folder|(sub)folder": Mailbox.
            date_range_start_time (datetime): Start date from where to stract data.
            date_range_end_time (datetime): End data up to where to stract data.
            limit (int, optional): Number of fetched top messages. Defaults to 10000.
            address_limit (int, optional): The maximum number of accepted characters in
                the sum of all email names. Defaults to 8000.
            outbox_list (List[str], optional): List of outbox folders to differenciate
                between Inboxes and Outboxes. Defaults to ["Sent Items"].

        Returns:
            list: A list with all messages from all Mailboxes.
        """
        if not outbox_list:
            outbox_list = ["Sent Items"]

        data = []
        for key, value in list(dict_folder.items()):
            count = 0
            for message in value.get_messages(limit=limit):
                received_time = message.received
                date_obj = datetime.fromisoformat(str(received_time))
                if (
                    date_range_start_time.replace(tzinfo=self.UTC)
                    < date_obj
                    < date_range_end_time.replace(tzinfo=self.UTC)
                ):
                    count += 1
                    fetched = message.to_api_data()
                    sender_mail = fetched.get("from", None)
                    if sender_mail is not None:
                        sender_mail = fetched["from"]["emailAddress"]["address"]
                    recivers_list = fetched.get("toRecipients")
                    recivers = " "

                    if recivers_list is not None:
                        for reciver in recivers_list:
                            add_string = f", {reciver['emailAddress']['address']}"
                            if (
                                sum(list(map(len, [recivers, add_string])))
                                >= address_limit
                            ):
                                break

                            recivers += add_string

                    categories = " "
                    if message.categories is not None:
                        categories = ", ".join(
                            categories for categories in message.categories
                        )

                    conversation_index = " "
                    if message.conversation_index is not None:
                        conversation_index = message.conversation_index

                    if isinstance(message.subject, str):
                        subject = message.subject.replace("\t", " ")
                    else:
                        subject = message.subject

                    row = {
                        "(sub)folder": value.name,
                        "conversation ID": fetched.get("conversationId"),
                        "conversation index": conversation_index,
                        "categories": categories,
                        "sender": sender_mail,
                        "subject": subject,
                        "recivers": recivers.strip(", "),
                        "received_time": fetched.get("receivedDateTime"),
                        "mail_adress": mailbox_name.split("@")[0]
                        .replace(".", "_")
                        .replace("-", "_"),
                    }
                    if any(x.lower() in key.lower() for x in outbox_list):
                        row["Inbox"] = False
                    else:
                        row["Inbox"] = True

                    data.append(row)

            if count > 0:
                self.logger.info(f"folder: {key.ljust(76, '-')}  messages: {count}")

        return data

    def api_connection(
        self,
        mailbox_name: str | None = None,
        request_retries: int = 10,
        start_date: str | None = None,
        end_date: str | None = None,
        limit: int = 10000,
        address_limit: int = 8000,
        outbox_list: list[str] | None = None,
    ) -> pd.DataFrame:
        """Download all the messages stored in a MailBox folder and subfolders.

        Args:
            mailbox_name (Optional[str], optional): Mailbox name. Defaults to None.
            request_retries (int, optional): How many times retries to authorizate.
                Defaults to 10.
            start_date (Optional[str], optional): A filtering start date parameter e.g.
                "2022-01-01". Defaults to None.
            end_date (Optional[str], optional): A filtering end date parameter e.g.
                "2022-01-02". Defaults to None.
            limit (int, optional): Number of fetched top messages. Defaults to 10000.
            address_limit (int, optional): The maximum number of accepted characters in
                the sum of all email names. Defaults to 8000.
            outbox_list (List[str], optional): List of outbox folders to differentiate
                between Inboxes and Outboxes. Defaults to ["Sent Items"].

        Returns:
            pd.DataFrame: All messages are stored in a pandas framework.
        """
        if not outbox_list:
            outbox_list = ["Sent Items"]
        account = Account(
            (self.credentials["client_id"], self.credentials["client_secret"]),
            auth_flow_type="credentials",
            tenant_id=self.credentials["tenant_id"],
            main_resource=mailbox_name,
            request_retries=request_retries,
        )

        if account.authenticate():
            self.logger.info(f"{mailbox_name} Authenticated!")
        else:
            msg = "Failed to authenticate."
            raise ValueError(msg)

        mailbox_obj = account.mailbox()

        if start_date is not None and end_date is not None:
            date_range_end_time = datetime.strptime(end_date, "%Y-%m-%d")
            date_range_start_time = datetime.strptime(start_date, "%Y-%m-%d")
        else:
            date_range_start_time = date.today() - timedelta(days=1)
            date_range_end_time = date.today()

            min_time = datetime.min.time()
            date_range_end_time = datetime.combine(date_range_end_time, min_time)
            date_range_start_time = datetime.combine(date_range_start_time, min_time)

        final_dict_folders = self._get_all_folders(mailbox_obj)

        self.data = self._get_messages_from_mailbox(
            mailbox_name=mailbox_name,
            dict_folder=final_dict_folders,
            date_range_start_time=date_range_start_time,
            date_range_end_time=date_range_end_time,
            limit=limit,
            address_limit=address_limit,
            outbox_list=outbox_list,
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
    ) -> pd.DataFrame:
        """Generate a pandas DataFrame with the data.

        Args:
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn

        Returns:
            pd.Dataframe: The response data as a pandas DataFrame plus viadot metadata.
        """
        super().to_df(if_empty=if_empty)

        data_frame = pd.DataFrame(self.data)

        if data_frame.empty:
            self._handle_if_empty(
                if_empty="warn",
                message="No data was got from the Mail Box for those days",
            )
        else:
            self.logger.info("Successfully downloaded data from the Mindful API.")

        return data_frame



import pytest

from viadot.sources.azure_sql import AzureSQL
from viadot.sources.sql_server import SQLServerCredentials


@pytest.fixture
def azure_sql_credentials():
    return SQLServerCredentials(
        user="test_user",
        password="test_password",  # pragma: allowlist secret # noqa: S106
        server="localhost",
        db_name="test_db",
        driver="ODBC Driver 17 for SQL Server",
    )


@pytest.fixture
def azure_sql(azure_sql_credentials: SQLServerCredentials, mocker):
    mocker.patch("viadot.sources.base.SQL.con", return_value=True)

    return AzureSQL(
        credentials={
            "user": azure_sql_credentials.user,
            "password": azure_sql_credentials.password,
            "server": azure_sql_credentials.server,
            "db_name": azure_sql_credentials.db_name,
            "data_source": "test_data_source",
        }
    )


def test_azure_sql_initialization(azure_sql):
    """Test that the AzureSQL object is initialized with the correct credentials."""
    assert azure_sql.credentials["server"] == "localhost"
    assert azure_sql.credentials["user"] == "test_user"
    assert (
        azure_sql.credentials["password"]
        == "test_password"  # pragma: allowlist secret  # noqa: S105
    )
    assert azure_sql.credentials["db_name"] == "test_db"


def test_create_external_database(azure_sql, mocker):
    """Test the create_external_database function."""
    mock_run = mocker.patch("viadot.sources.base.SQL.run", return_value=True)

    # Test parameters
    external_database_name = "test_external_db"
    storage_account_name = "test_storage_account"
    container_name = "test_container"
    sas_token = "test_sas_token"  # noqa: S105
    master_key_password = (
        "test_master_key_password"  # pragma: allowlist secret # noqa: S105
    )
    credential_name = "custom_credential_name"

    azure_sql.create_external_database(
        external_database_name=external_database_name,
        storage_account_name=storage_account_name,
        container_name=container_name,
        sas_token=sas_token,
        master_key_password=master_key_password,
        credential_name=credential_name,
    )

    # Expected SQL commands with custom credential name
    expected_master_key_sql = (
        f"CREATE MASTER KEY ENCRYPTION BY PASSWORD = {master_key_password}"
    )

    mock_run.assert_any_call(expected_master_key_sql)
    assert mock_run.call_count == 3  # Ensure all 3 SQL commands were executed



from collections import OrderedDict

from pandas import DataFrame
import pendulum
import pytest

from viadot.utils import skip_test_on_missing_extra


try:
    from viadot.sources import SAPRFC
except ImportError:
    skip_test_on_missing_extra(source_name="SAPRFC", extra="sap")

credentials = {"a": "b"}

sap = SAPRFC(credentials=credentials)

sql1 = "SELECT a AS a_renamed, b FROM table1 WHERE table1.c = 1"
sql2 = "SELECT a FROM fake_schema.fake_table WHERE a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2) LIMIT 5 OFFSET 3"
sql3 = "SELECT b FROM c WHERE testORword=1 AND testANDword=2 AND testLIMITword=3 AND testOFFSETword=4"
sql4 = "SELECT c FROM d WHERE testLIMIT = 1 AND testOFFSET = 2 AND LIMITtest=3 AND OFFSETtest=4"
sql5 = sql3 + " AND longword123=5"
sql6 = "SELECT a FROM fake_schema.fake_table WHERE a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2) AND longcolname=3 AND otherlongcolname=5 LIMIT 5 OFFSET 3"
sql7 = """
SELECT a, b
FROM b
WHERE c = 1
AND d = 2
AND longcolname = 12345
AND otherlongcolname = 6789
AND thirdlongcolname = 01234
LIMIT 5
OFFSET 10
"""


def test__get_table_name():
    assert sap._get_table_name(sql1) == "table1"
    assert sap._get_table_name(sql2) == "fake_schema.fake_table", sap._get_table_name(
        sql2
    )
    assert sap._get_table_name(sql7) == "b"


def test__get_columns():
    assert sap._get_columns(sql1) == ["a", "b"]
    assert sap._get_columns(sql1, aliased=True) == ["a_renamed", "b"], sap._get_columns(
        sql1, aliased=True
    )
    assert sap._get_columns(sql2) == ["a"]
    assert sap._get_columns(sql7) == ["a", "b"]


def test__get_where_condition():
    assert sap._get_where_condition(sql1) == "table1.c = 1", sap._get_where_condition(
        sql1
    )
    assert (
        sap._get_where_condition(sql2) == "a=1 AND b=2 OR c LIKE 'a%' AND d IN (1, 2)"
    ), sap._get_where_condition(sql2)
    assert (
        sap._get_where_condition(sql3)
        == "testORword=1 AND testANDword=2 AND testLIMITword=3 AND testOFFSETword=4"
    ), sap._get_where_condition(sql3)
    assert (
        sap._get_where_condition(sql4)
        == "testLIMIT = 1 AND testOFFSET = 2 AND LIMITtest=3 AND OFFSETtest=4"
    ), sap._get_where_condition(sql4)
    assert (
        sap._get_where_condition(sql7)
        == "c = 1 AND d = 2 AND longcolname = 12345 AND otherlongcolname = 6789"
    ), sap._get_where_condition(sql7)


def test__get_limit():
    assert sap._get_limit(sql1) is None
    assert sap._get_limit(sql2) == 5
    assert sap._get_limit(sql7) == 5


def test__get_offset():
    assert sap._get_offset(sql1) is None
    assert sap._get_offset(sql2) == 3
    assert sap._get_offset(sql7) == 10


def test_client_side_filters_simple():
    _ = sap._get_where_condition(sql5)
    assert sap.client_side_filters == OrderedDict(
        {"AND": "longword123=5"}
    ), sap.client_side_filters


def test_client_side_filters_with_limit_offset():
    _ = sap._get_where_condition(sql6)
    assert sap.client_side_filters == OrderedDict(
        {"AND": "otherlongcolname=5"}
    ), sap.client_side_filters

    _ = sap._get_where_condition(sql7)
    assert sap.client_side_filters == OrderedDict(
        {"AND": "thirdlongcolname = 01234"}
    ), sap.client_side_filters


def test___build_pandas_filter_query():
    _ = sap._get_where_condition(sql6)
    assert (
        sap._build_pandas_filter_query(sap.client_side_filters)
        == "otherlongcolname == 5"
    ), sap._build_pandas_filter_query(sap.client_side_filters)
    _ = sap._get_where_condition(sql7)
    assert (
        sap._build_pandas_filter_query(sap.client_side_filters)
        == "thirdlongcolname == 01234"
    ), sap._build_pandas_filter_query(sap.client_side_filters)


def test__adjust_whitespaces():
    sap.rfc_unique_id = ["column1", "column2"]
    sap._rfc_unique_id_len = {"column1": 5, "column2": 4}
    data = {"column1": ["xyz  ", "oiu"], "column2": ["yrt ", "lkj"]}
    df = DataFrame(data)
    df = sap._adjust_whitespaces(df)
    col_values_len = df.applymap(lambda x: len(x))
    check_if_length_match = col_values_len == sap._rfc_unique_id_len.values()
    assert check_if_length_match.all().all()


def test_parse_dates_success():
    """Test `_parse_dates` function."""
    query = "SELECT * FROM table WHERE date_column = <<today>>"
    result = sap._parse_dates(query)
    assert (
        result
        == f"SELECT * FROM table WHERE date_column = {pendulum.today().strftime('%Y%m%d')}"  # noqa: S608
    )


def test_parse_dates_raises_typeerror():
    """Test `_parse_dates` for not supported dynamic date type."""
    query = "SELECT * FROM table WHERE date_column in (<<last_3_years>>)"

    with pytest.raises(TypeError):
        sap._parse_dates(query)



"""Tasks for interacting with [Luma](https://github.com/dyvenia/luma)."""

import logging
import os
from pathlib import Path
from typing import Any, Literal

from prefect import task
from prefect.logging import get_run_logger

from viadot.orchestration.prefect.utils import shell_run_command


@task(retries=2, retry_delay_seconds=5, timeout_seconds=60 * 10)
async def luma_ingest_task(  # noqa: PLR0913
    metadata_dir_path: str | Path,
    luma_url: str = "http://localhost:8000",
    metadata_kind: Literal["model", "model_run"] = "model",
    follow: bool = False,
    env: dict[str, Any] | None = None,
    shell: str = "bash",
    return_all: bool = True,
    stream_level: int = logging.INFO,
    max_retries: int = 3,  # noqa: ARG001
    logger: logging.Logger | None = None,
    raise_on_failure: bool = True,
) -> list[str]:
    """Runs Luma ingestion by sending dbt artifacts to Luma ingestion API.

    Args:
        metadata_dir_path: The path to the directory containing metadata files.
            In the case of dbt, it's dbt project's `target` directory,
            which contains dbt artifacts (`sources.json`, `catalog.json`,
            `manifest.json`, and `run_results.json`).
        luma_url: The URL of the Luma instance to ingest into.
        metadata_kind: The kind of metadata to ingest. Either `model` or `model_run`.
        follow: Whether to follow the ingestion process until it's completed (by
            default, ingestion request is sent without awaiting for the response). By
            default, `False`.
        env: Dictionary of environment variables to use for
            the subprocess; can also be provided at runtime.
        shell: Shell to run the command with.
        return_all: Whether this task should return all lines of stdout as a list,
            or just the last line as a string.
        stream_level: The logging level of the stream;
            defaults to 20; equivalent to `logging.INFO`.
        max_retries: The maximum number of times to retry the task. Defaults to 3.
        logger: The logger to use for logging the task's output. By default, the
            Prefect's task run logger.
        raise_on_failure: Whether to raise an exception if the command fails.

    Returns:
        list[str]: Lines from stdout as a list.

    Example:
        ```python
        from prefect import flow
        from viadot.tasks import luma_ingest_task

        metadata_dir_path = "${HOME}/dbt/my_dbt_project/target"

        @flow
        def example_luma_ingest_flow():
            return luma_ingest_task(metadata_dir_path=metadata_dir_path)

        example_luma_ingest_flow()
        ```
    """
    if not logger:
        logger = get_run_logger()

    if isinstance(metadata_dir_path, str):
        path_expanded = os.path.expandvars(metadata_dir_path)
        metadata_dir_path = Path(path_expanded)

    luma_command = "ingest" if metadata_kind == "model" else "send-test-results"
    follow_flag = "--follow" if follow else ""
    command = (
        f"luma dbt {luma_command} -m {metadata_dir_path} -l {luma_url} {follow_flag}"
    )

    return await shell_run_command(
        command=command,
        env=env,
        shell=shell,
        return_all=return_all,
        stream_level=stream_level,
        logger=logger,
        raise_on_failure=raise_on_failure,
    )



"""Tasks for interacting with [dbt](https://www.getdbt.com/)."""

import logging
import os
from typing import Any

from prefect import task
from prefect.logging import get_run_logger

from viadot.orchestration.prefect.utils import shell_run_command


@task(retries=0, timeout_seconds=2 * 60 * 60)
async def dbt_task(
    command: str = "run",
    project_path: str | None = None,
    env: dict[str, Any] | None = None,
    shell: str = "bash",
    return_all: bool = False,
    stream_level: int = logging.INFO,
    raise_on_failure: bool = True,
) -> list[str] | str:
    """Runs dbt commands within a shell.

    Args:
        command: dbt command to be executed; can also be provided post-initialization
            by calling this task instance.
        project_path: The path to the dbt project.
        env: Dictionary of environment variables to use for the subprocess; can also be
            provided at runtime.
        shell: Shell to run the command with.
        return_all: Whether this task should return all lines of stdout as a list, or
            just the last line as a string.
        stream_level: The logging level of the stream; defaults to 20, equivalent to
            `logging.INFO`.
        raise_on_failure: Whether to fail the task if the command fails.

    Returns:
        If return all, returns all lines as a list; else the last line as a string.

    Example:
        Executes `dbt run` on a specified dbt project.
        ```python
        from prefect import flow
        from viadot.tasks import dbt_task

        PROJECT_PATH = "/home/viadot/dbt/my_dbt_project"

        @flow
        def example_dbt_task_flow():
            return dbt_task(
                command="run", project_path=PROJECT_PATH, return_all=True
            )

        example_dbt_task_flow()
        ```
    """
    logger = get_run_logger()

    project_path = os.path.expandvars(project_path) if project_path is not None else "."

    return await shell_run_command(
        command=f"dbt {command}",
        env=env,
        helper_command=f"cd {project_path}",
        shell=shell,
        return_all=return_all,
        stream_level=stream_level,
        raise_on_failure=raise_on_failure,
        logger=logger,
    )



import pandas as pd
import pytest

from src.viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from src.viadot.orchestration.prefect.tasks import vid_club_to_df


EXPECTED_DF = pd.DataFrame(
    {"id": [1, 2], "name": ["Company A", "Company B"], "region": ["pl", "ro"]}
)


class MockVidClub:
    def __init__(self, *args, **kwargs):
        """Init method."""
        pass

    def to_df(self):
        return EXPECTED_DF


def test_vid_club_to_df(mocker):
    mocker.patch("viadot.orchestration.prefect.tasks.VidClub", new=MockVidClub)

    df = vid_club_to_df(
        endpoint="company",
        from_date="2023-01-01",
        to_date="2023-12-31",
        items_per_page=100,
        region="pl",
        vidclub_credentials_secret="VIDCLUB",  # pragma: allowlist secret # noqa: S106
    )

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert df.equals(EXPECTED_DF)


def test_vid_club_to_df_missing_credentials(mocker):
    mocker.patch(
        "viadot.orchestration.prefect.tasks.get_credentials", return_value=None
    )

    with pytest.raises(MissingSourceCredentialsError):
        vid_club_to_df(
            endpoint="company",
            from_date="2023-01-01",
            to_date="2023-12-31",
            items_per_page=100,
            region="pl",
            vidclub_credentials_secret="VIDCLUB",  # pragma: allowlist secret # noqa: S106
        )



"""Download data from Matomo API and load into Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow
from prefect.task_runners import ConcurrentTaskRunner

from viadot.orchestration.prefect.tasks import df_to_redshift_spectrum, matomo_to_df


@flow(
    name="Matomo extraction to Redshift Spectrum",
    description="Extract data from Matomo API and load into Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
    task_runner=ConcurrentTaskRunner,
)
def matomo_to_redshift_spectrum(  # noqa: PLR0913
    matomo_url: str,
    top_level_fields: list[str],
    record_path: str | list[str],
    to_path: str,
    schema_name: str,
    table: str,
    params: dict[str, Any],
    matomo_config_key: str | None = None,
    matomo_credentials_secret: str | None = None,
    record_prefix: str | None = None,
    if_empty: Literal["warn", "skip", "fail"] = "warn",
    tests: dict[str, Any] | None = None,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    compression: Literal["snappy", "gzip", "zstd"] | None = None,
    aws_sep: str = ",",
    credentials_secret: str | None = None,
    aws_config_key: str | None = None,
) -> None:
    """Flow for downloading data from Matomo to Redshift Spectrum.

    Args:
        matomo_url (str): The base matomo url of the Matomo instance.
        top_level_fields (list[str]): List of top level fields to get from the API JSON.
        record_path (str | list[str]): The path field to the records
            in the API response.
            Could be handled as a list of path + fields to extract:
                    record_path = 'actionDetails'
                    record_path = ['actionDetails', 'eventAction']
        to_path (str): Path to a S3 folder where the table will be located.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        matomo_config_key (Optional[str], optional): The key in the viadot config
             holding relevant Matomo credentials. Defaults to None.
        matomo_credentials_secret (str, optional): The name of the secret that stores
            Matomo credentials. Defaults to None.
        params (dict[str, Any]): Parameters for the API request.
                Required params: "module","method","idSite","period","date","format".
        record_prefix (Optional[str], optional): A prefix for the record path fields.
            For example: "action_". Defaults to None.
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            query returns no data. Defaults to "warn".
        tests (dict[str, Any], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from viadot.utils. Defaults to None.
        extension (str): Required file type. Accepted file formats are 'csv' and
            'parquet'. Defaults to ".parquet".
        if_exists (Literal["overwrite", "append"]): Whether to overwrite or append to
            existing data. Defaults to "overwrite".
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Defaults to None.
        compression (Literal["snappy", "gzip", "zstd"], optional): Compression style
            (None, snappy, gzip, zstd). Defaults to None.
        aws_sep (str, optional): Field delimiter for the output file. Defaults to ','.
        credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        aws_config_key (str, optional): The key in the viadot config holding relevant
            AWS credentials. Defaults to None.

    Examples:
        matomo_to_redshift_spectrum(
            matomo_url="https://matomo.example.com",
            top_level_fields=["idSite", "visitorId", "visitIp"],
            record_path="actionDetails",
            to_path="s3://your-bucket/matomo/",
            schema_name="staging",
            table="matomo_visits",
            matomo_config_key="matomo_prod",
            params={
                "module": "API",
                "method": "Live.getLastVisitsDetails",
                "idSite": "53",
                "period": "range",
                "date": ("<<yesterday>>", "<<yesterday>>"),
                "format": "JSON"
            },
            tests={
                "unique": ["idsite", "visitIp"],
                "not_null": ["serverDate"],
            },
            partition_cols=["serverDate"],
            compression="snappy",
            if_exists="overwrite",
        )
    """
    data_frame = matomo_to_df(
        url=matomo_url,
        top_level_fields=top_level_fields,
        record_path=record_path,
        config_key=matomo_config_key,
        credentials_secret=matomo_credentials_secret,
        params=params,
        record_prefix=record_prefix,
        if_empty=if_empty,
        tests=tests,
    )

    return df_to_redshift_spectrum(
        df=data_frame,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        compression=compression,
        sep=aws_sep,
        config_key=aws_config_key,
        credentials_secret=credentials_secret,
    )



"""Tasks for interacting with git repositories."""

import logging
import shutil
from typing import Any

from prefect import get_run_logger, task
import pygit2

from viadot.orchestration.prefect.utils import get_credentials


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 10)
def clone_repo(
    url: str,
    token: str | None = None,
    token_secret: str | None = None,
    logger: logging.Logger | None = None,
    **kwargs: dict[str, Any] | None,
) -> None:
    """Clone Azure DevOps or GitHub repository.

    Args:
        url (str): Alternatively to URL + token or URL + token_secret, you can also
            provide the full URL (including token, if the repo is private) here.
        token (str, optional): The token to use to clone the repo.
        token_secret (str, optional): The secret holding the token.
        logger (logging.Logger): The logger to use. By default, Prefect's task run
            logger is used.
        token (str, optional): The personal access token.
        token_secret (str, optional): The name of the secret storing the token. Defaults
            to None.
        logger (logging.Logger, optional): The logger to use for logging the task's
            output. By default, Prefect's task run logger.
        kwargs (dict): Keyword arguments to be passed to `pygit2.clone_repository()`.
        **kwargs (dict, optional): Keyword arguments to pass to
            `pygit2.clone_repository()`.

    Examples:
        Azure DevOps (public repo):
        https://dev.azure.com/{organization_name}/{project_name}/_git/{repo_name}
        Azure DevOps (private repo):
        https://{token}@dev.azure.com/{organization_name}/{project_name}/_git/{repo_name}
        GitHub (public repo): https://github.com/{organization_name}/{repo_name}.git
        GitHub (private repo): https://{token}@github.com/{organization_name}/{repo_name}.git
    """
    if not logger:
        logger = get_run_logger()

    url = url.strip("/")

    if token_secret:
        token = get_credentials(token_secret)

    if token:
        url = url.replace("https://dev.azure.com", f"https://{token}@dev.azure.com")
        url = url.replace("https://github.com", f"https://{token}@github.com")
        url = url.replace("gitlab", f"oauth2:{token}@gitlab")

    repo_name = url.split("/")[-1].replace(".git", "")
    path = kwargs.get("path") or repo_name
    kwargs["path"] = path

    logger.info(f"Removing {path}...")
    shutil.rmtree(path, ignore_errors=True)  # Delete folder on run

    logger.info(f"Cloning repo '{repo_name}' into {path}...")
    pygit2.clone_repository(url, **kwargs)
    logger.info(f"Repo '{repo_name}' has been successfully cloned into {path}.")



"""Task for downloading data from Epicor Prelude API."""

from typing import Literal

import pandas as pd
from prefect import task
from prefect.logging import get_run_logger

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources.epicor import Epicor


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60 * 3)
def epicor_to_df(
    base_url: str,
    filters_xml: str,
    validate_date_filter: bool = True,
    start_date_field: str = "BegInvoiceDate",
    end_date_field: str = "EndInvoiceDate",
    credentials_secret: str | None = None,
    config_key: str | None = None,
    if_empty: Literal["warn", "fail", "skip"] = "warn",
) -> pd.DataFrame:
    """Load the data from Epicor Prelude API into a pandas DataFrame.

    Args:
        base_url (str, required): Base url to Epicor.
        filters_xml (str, required): Filters in form of XML. The date filter
             is required.
        validate_date_filter (bool, optional): Whether or not validate xml date filters.
                Defaults to True.
        start_date_field (str, optional) The name of filters field containing
            start date. Defaults to "BegInvoiceDate".
        end_date_field (str, optional) The name of filters field containing end date.
                Defaults to "EndInvoiceDate".
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        if_empty (Literal["warn", "skip", "fail"], optional): Action to take if
                the DataFrame is empty.
                - "warn": Logs a warning.
                - "skip": Skips the operation.
                - "fail": Raises an error.
                Defaults to "warn".

    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )
    epicor = Epicor(
        credentials=credentials,
        base_url=base_url,
        validate_date_filter=validate_date_filter,
        start_date_field=start_date_field,
        end_date_field=end_date_field,
    )
    df = epicor.to_df(filters_xml=filters_xml, if_empty=if_empty)  # type: ignore
    nrows = df.shape[0]
    ncols = df.shape[1]

    logger.info(
        f"Successfully downloaded {nrows} rows and {ncols} columns of data to a DataFrame."
    )
    return df



"""Mindful API connector."""

from datetime import date, timedelta
from io import StringIO
import json
from typing import Any, Literal

import pandas as pd
from pydantic import BaseModel
from requests.auth import HTTPBasicAuth
from requests.models import Response

from viadot.config import get_source_credentials
from viadot.exceptions import APIError, CredentialError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_response


class MindfulCredentials(BaseModel):
    """Checking for values in Mindful credentials dictionary.

    Two key values are held in the Mindful connector:
        - customer_uuid: The unique ID for the organization.
        - auth_token: A unique token to be used as the password for API requests.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    customer_uuid: str
    auth_token: str


class Mindful(Source):
    """Class implementing the Mindful API.

    Documentation for this API is available at: https://apidocs.surveydynamix.com/.
    """

    ENDPOINTS = ("interactions", "responses", "surveys")

    def __init__(
        self,
        *args,
        credentials: MindfulCredentials | None = None,
        config_key: str = "mindful",
        region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
        **kwargs,
    ):
        """Create a Mindful instance.

        Args:
            credentials (Optional[MindfulCredentials], optional): Mindful credentials.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "mindful".
            region (Literal[us1, us2, us3, ca1, eu1, au1], optional): Survey Dynamix
                region from where to interact with the mindful API. Defaults to "eu1"
                English (United Kingdom).

        Examples:
            mindful = Mindful(
                credentials=credentials,
                config_key=config_key,
                region=region,
            )
            mindful.api_connection(
                endpoint=endpoint,
                date_interval=date_interval,
                limit=limit,
            )
            data_frame = mindful.to_df()
        """
        credentials = credentials or get_source_credentials(config_key) or None
        if credentials is None:
            msg = "Missing credentials."
            raise CredentialError(msg)

        validated_creds = dict(MindfulCredentials(**credentials))
        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.auth = (credentials["customer_uuid"], credentials["auth_token"])
        if region != "us1":
            self.region = f"{region}."
        else:
            self.region = ""

    def _mindful_api_response(
        self,
        params: dict[str, Any] | None = None,
        endpoint: str = "",
    ) -> Response:
        """Call to Mindful API given an endpoint.

        Args:
            params (Optional[Dict[str, Any]], optional): Parameters to be passed into
                the request. Defaults to None.
            endpoint (str, optional): API endpoint for an individual request.
                Defaults to "".

        Returns:
            Response: request object with the response from the Mindful API.
        """
        return handle_api_response(
            url=f"https://{self.region}surveydynamix.com/api/{endpoint}",
            params=params,
            method="GET",
            auth=HTTPBasicAuth(*self.auth),
        )

    def api_connection(
        self,
        endpoint: Literal["interactions", "responses", "surveys"] = "surveys",
        date_interval: list[date] | None = None,
        limit: int = 1000,
    ) -> None:
        """General method to connect to Survey Dynamix API and generate the response.

        Args:
            endpoint (Literal["interactions", "responses", "surveys"], optional): API
                endpoint for an individual request. Defaults to "surveys".
            date_interval (Optional[List[date]], optional): Date time range detailing
                the starting date and the ending date. If no range is passed, one day of
                data since this moment will be retrieved. Defaults to None.
            limit (int, optional): The number of matching interactions to return.
                Defaults to 1000.

        Raises:
            ValueError: Not available endpoint.
            APIError: Failed to download data from the endpoint.
        """
        if endpoint not in self.ENDPOINTS:
            raise ValueError(
                f"Survey Dynamix endpoint: '{endpoint}',"
                + " is not available through Mindful viadot connector."
            )

        if (
            date_interval is None
            or all(list(map(isinstance, date_interval, [date] * len(date_interval))))
            is False
        ):
            reference_date = date.today()
            date_interval = [reference_date - timedelta(days=1), reference_date]

            self.logger.warning(
                "No `date_interval` parameter was defined, or was erroneously "
                + "defined. `date_interval` parameter must have the folloing "
                + "structure:\n\t[`date_0`, `date_1`], having that `date_1` > "
                + "`date_0`.\nBy default, one day of data, from "
                + f"{date_interval[0].strftime('%Y-%m-%d')} to "
                + f"{date_interval[1].strftime('%Y-%m-%d')}, will be obtained."
            )

        params = {
            "_limit": limit,
            "start_date": f"{date_interval[0]}",
            "end_date": f"{date_interval[1]}",
        }

        if endpoint == "surveys":
            del params["start_date"]
            del params["end_date"]

        response = self._mindful_api_response(
            endpoint=endpoint,
            params=params,
        )
        response_ok = 200
        no_data_code = 204
        if response.status_code == response_ok:
            self.logger.info(
                f"Successfully downloaded '{endpoint}' data from mindful API."
            )
            self.data = StringIO(response.content.decode("utf-8"))
        elif response.status_code == no_data_code and not response.content.decode():
            self.logger.warning(
                f"There are not '{endpoint}' data to download from"
                + f" {date_interval[0]} to {date_interval[1]}."
            )
            self.data = json.dumps({})
        else:
            self.logger.error(
                f"Failed to downloaded '{endpoint}' data. - {response.content}"
            )
            msg = f"Failed to downloaded '{endpoint}' data."
            raise APIError(msg)

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
    ) -> pd.DataFrame:
        """Download the data to a pandas DataFrame.

        Args:
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn

        Returns:
            pd.Dataframe: The response data as a pandas DataFrame plus viadot metadata.
        """
        super().to_df(if_empty=if_empty)

        data_frame = pd.read_json(self.data)

        if data_frame.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info("Successfully downloaded data from the Mindful API.")

        return data_frame



"""Source for ingesting data from European Central Bank (ECB) exchange rates API."""

from typing import Any, Literal

from defusedxml import ElementTree as ElementDE
import pandas as pd

from viadot.exceptions import APIError
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    handle_api_response,
    validate,
)


class ECBExchangeRates(Source):
    """ECB exchange rates source class for fetching daily exchange rates from ECB."""

    URL = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-daily.xml"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        """Initialize ECB exchange rates source.

        Connector allows to pull daily exchange rates data from ECB API and convert to
        pandas DataFrame. No credentials are required.

        Usage:
            ecb = ECBExchangeRates()
            df = ecb.to_df()

        Example:
            ecb = ECBExchangeRates()
            df = ecb.to_df()
        """
        super().__init__(*args, credentials=None, **kwargs)

    def fetch_data(self) -> str:
        """Connect to ECB API and fetch XML response data.

        The function sends a GET request to the ECB exchange rates endpoint.
        If the request is successful, the response XML is returned as a string.
        If the request fails, an APIError is raised with the error message.

        Returns:
            str: The API response data as XML string.

        Raises:
            APIError: If the request fails.
        """
        try:
            response = handle_api_response(
                url=self.URL,
                params=None,
                method="GET",
            )
        except Exception as e:
            msg = "Failed to fetch data from ECB API"
            self.logger.exception(msg)
            raise APIError(msg) from e

        self.logger.info("Successfully fetched data from ECB API.")
        return response.text

    def _parse_xml(self, xml_string: str) -> pd.DataFrame:
        """Parse XML string and extract exchange rates data.

        Args:
            xml_string (str): The XML response string from ECB API.

        Returns:
            pd.DataFrame: DataFrame containing time, currency, and rate columns.

        Raises:
            ValueError: If the XML cannot be parsed or has unexpected structure.
        """
        try:
            root = ElementDE.fromstring(xml_string)
        except Exception as e:
            msg = "Failed to parse XML response from ECB API"
            self.logger.exception(msg)
            raise ValueError(msg) from e

        # Define namespaces - ECB uses default namespace for Cube elements
        # We need to register the namespace or use the full namespace URL
        namespace = "http://www.ecb.int/vocabulary/2002-08-01/eurofxref"

        # Find all Cube elements with currency and rate attributes
        data = []
        cube_elements = root.findall(f".//{{{namespace}}}Cube[@currency]")

        # Get the time from the parent Cube element
        time_cube = root.find(f".//{{{namespace}}}Cube[@time]")
        if time_cube is None:
            msg = "Could not find time attribute in XML response"
            self.logger.error(msg)
            raise ValueError(msg)

        time_value = time_cube.get("time")

        # Extract currency and rate from each Cube element
        for cube in cube_elements:
            currency = cube.get("currency")
            rate = cube.get("rate")

            if currency and rate:
                data.append(
                    {
                        "time": time_value,
                        "currency": currency,
                        "rate": float(rate),
                    }
                )

        if not data:
            self.logger.warning("No exchange rate data found in XML response.")
            return pd.DataFrame(columns=["time", "currency", "rate"])

        df = pd.DataFrame(data)
        self.logger.info(f"Successfully parsed {len(df)} exchange rates from ECB data.")
        return df

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        tests: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Convert ECB exchange rates data to pandas DataFrame.

        Args:
            if_empty (Literal["warn", "skip", "fail"], optional): What to do if no data
                is available. Defaults to "warn".
            tests (dict[str, Any], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame containing exchange rates data with columns:
                - time: The date of the exchange rates
                - currency: The currency code (e.g., USD, GBP)
                - rate: The exchange rate against EUR

        Raises:
            ValueError: If no data has been fetched or if validation fails.
        """
        self.logger.info("Converting ECB exchange rates data to pandas DataFrame.")

        # Fetch XML data
        xml_data = self.fetch_data()

        # Parse XML to DataFrame
        df = self._parse_xml(xml_data)

        if df.empty:
            self.logger.warning("No exchange rates found in the response.")
            self._handle_if_empty(if_empty=if_empty)

        if tests:
            self.logger.info("Running data validation tests.")
            validate(df=df, tests=tests)

        self.logger.info(
            f"Successfully processed {len(df)} exchange rates from ECB data."
        )

        return df



"""Build specified dbt model(s) and upload the generated metadata to Luma."""

from datetime import datetime, timezone
from pathlib import Path
import re
import shutil
from typing import Literal

from prefect import flow, task
from prefect.logging import get_run_logger
from prefect.states import Failed

from viadot.orchestration.prefect.tasks import (
    clone_repo,
    dbt_task,
    luma_ingest_task,
    s3_upload_file,
)
from viadot.orchestration.prefect.utils import get_credentials


@task
def remove_dbt_repo_dir(dbt_repo_dir_name: str) -> None:
    """Remove the repo directory.

    Args:
        dbt_repo_dir_name (str): The name of the dbt repo directory.
    """
    shutil.rmtree(dbt_repo_dir_name, ignore_errors=True)


@flow(
    name="Transform and Catalog",
    description="Build specified dbt model(s) and upload generated metadata to Luma.",
    timeout_seconds=2 * 60 * 60,
)
def transform_and_catalog(  # noqa: PLR0912, PLR0913, PLR0915, C901
    dbt_repo_url: str | None = None,
    dbt_repo_url_secret: str | None = None,
    dbt_project_path: str = "dbt",
    dbt_repo_branch: str | None = None,
    dbt_repo_token_secret: str | None = None,
    dbt_selects: dict[str, str] | None = None,
    dbt_target: str | None = None,
    dbt_target_dir_path: str | None = None,
    luma_url: str = "http://localhost:8000",
    luma_follow: bool = False,
    metadata_kind: Literal["model", "model_run"] = "model_run",
    run_results_storage_path: str | None = None,
    run_results_storage_config_key: str | None = None,
    run_results_storage_credentials_secret: str | None = None,
    fail_flow_only_on_build_failure: bool = False,
) -> list[str]:
    """Build specified dbt model(s) and upload the generated metadata to Luma.

    Supports ingesting both model and model run metadata (controlled by the
    `metadata_kind` parameter).

    Note that metadata is still ingested even if the preceding `dbt test` task fails.
    This is done in order to capture test failure metadata in the data catalog.

    Args:
        dbt_repo_url (str, optional): The URL for cloning the dbt repo with relevant
            dbt project. Defaults to None.
        dbt_repo_url_secret (str, optional): Alternatively to above, the secret
            containing `dbt_repo_url`. Defaults to None.
        dbt_project_path (str): Path to the dbt project directory, relative to the
            dbt repository's root. For example, "dbt/my_dbt_project". Defaults to "dbt".
        dbt_repo_branch (str, optional): The branch of the dbt repo to use. Defaults to
            None (default repo branch).
        dbt_repo_token_secret (str, optional): The secret containing the personal access
            token used to clone the dbt repository, in case it's private. Not required
            if token is already included in `dbt_repo_url` (which is NOT recommended).
            Defaults to None.
        dbt_selects (dict, optional): Valid
            [dbt node selection](https://docs.getdbt.com/reference/node-selection/syntax)
            expressions. Valid keys are `run`, `test`,`build`, and `source_freshness`.
                The test select expression is taken from run's, as long as run select is
                provided. Defaults to None.
        dbt_target (str): The dbt target to use. If not specified, the default dbt
            target (as specified in `profiles.yaml`) will be used. Defaults to None.
        dbt_target_dir_path (str): The path to your dbt project's target
            directory, which contains dbt artifact JSON files, relative
            to dbt project's root directory. By default,
            `<repo_name>/<dbt_project_path>/target`, since "target" is the default
            name of the directory generated by dbt.
        luma_url (str, optional): The URL of the Luma instance to ingest into.
            Defaults to "http://localhost:8000".
        luma_follow (bool, optional): Whether to follow the ingestion process until it's
            completed (by default, ingestion request is sent without awaiting for the
            response). By default, `False`.
        metadata_kind (Literal["model", "model_run"], optional): The kind of metadata
            to ingest. Defaults to "model_run".
        run_results_storage_path (str, optional): The directory to upload the
            `run_results.json` file to. Note that a timestamp will be appended to the
            end of the file. Currently, only S3 is supported. Defaults to None.
        run_results_storage_config_key (str, optional): The key in the viadot config
            holding AWS credentials. Defaults to None.
        run_results_storage_credentials_secret (str, optional): The name of the secret
            block in Prefect holding AWS credentials. Defaults to None.
        fail_flow_only_on_build_failure (bool): Determines the flow's failure behavior
            based on dbt build outcomes.
            When False (default):
                - The flow will fail on any dbt build failure (including test failures)
            When True:
                - The flow will only fail if model building fails
                - Test failures alone won't cause the flow failure

    Returns:
        list[str]: Lines from stdout of the `upload_metadata` task as a list.

    Examples:
        # Build staging models.

        ```python
        import os
        from prefect_viadot.flows import transform_and_catalog

        my_dbt_repo_url = "https://github.com/dbt-labs/jaffle_shop"
        my_luma_url = "http://localhost:8000"

        transform_and_catalog(
            dbt_repo_url=my_dbt_repo_url
            dbt_selects={"run": "staging"}
            luma_url=my_luma_url,
            run_results_storage_path="s3://my-bucket/dbt/run_results",
            run_results_storage_credentials_secret="my-aws-credentials-block",
        )
        ```

        Some common `dbt_selects` patterns:
        - runs a specific model and all its downstream dependencies:
            `dbt_select={"run": "my_model+"}`
        - runs all models in a directory:
            `dbt_select={"run: "models/staging"}`
        - runs a specific model in a folder:
            `dbt_select={"run": "marts.domain.some_model"}`
        - runs tests for a specific model:
            `dbt_select={"test": "my_model"}`
        - build a specific model:
            `dbt_select={"build": "my_model"}`
        - build all models in a folder:
            `dbt_select={"build": "models.intermediate"}`
    """
    logger = get_run_logger()
    # Clone the dbt project.
    dbt_repo_url = dbt_repo_url or get_credentials(dbt_repo_url_secret)
    clone = clone_repo(
        url=dbt_repo_url,
        checkout_branch=dbt_repo_branch,
        token_secret=dbt_repo_token_secret,
    )

    # Prepare the environment.
    dbt_repo_name = dbt_repo_url.split("/")[-1].replace(".git", "")
    dbt_project_path_full = Path(dbt_repo_name) / dbt_project_path
    dbt_pull_deps_task = dbt_task.with_options(
        name="dbt_deps", retries=3, retry_delay_seconds=60
    )
    pull_dbt_deps = dbt_pull_deps_task(
        project_path=dbt_project_path_full,
        command="deps",
        wait_for=[clone],
    )

    # Run dbt commands.
    dbt_target_option = f"-t {dbt_target}" if dbt_target is not None else ""

    task_failed = False

    if metadata_kind == "model_run":
        # Produce `run-results.json` artifact for Luma ingestion.
        if dbt_selects:
            build_select = dbt_selects.get("build")
            run_select = dbt_selects.get("run")
            test_select = dbt_selects.get("test", run_select)

            build_select_safe = f"-s {build_select}" if build_select is not None else ""
            run_select_safe = f"-s {run_select}" if run_select is not None else ""
            test_select_safe = f"-s {test_select}" if test_select is not None else ""
        else:
            run_select_safe = ""
            test_select_safe = ""
        if build_select:
            # If build task is used, run and test tasks are not needed.
            # Build task executes run and tests commands internally.
            build_task = dbt_task.with_options(name="dbt_build")
            raise_on_failure = not fail_flow_only_on_build_failure
            try:
                build = build_task(
                    project_path=dbt_project_path_full,
                    command=f"build {build_select_safe} {dbt_target_option}",
                    wait_for=[pull_dbt_deps],
                    raise_on_failure=raise_on_failure,
                    return_all=True,
                )
            except Exception:
                msg = "Build task failed."
                logger.exception(msg)
                build = msg
                task_failed = True

            upload_metadata_upstream_task = build
        else:
            run_task = dbt_task.with_options(name="dbt_run")
            run = run_task(
                project_path=dbt_project_path_full,
                command=f"run {run_select_safe} {dbt_target_option}",
                wait_for=[pull_dbt_deps],
            )

            test_task = dbt_task.with_options(name="dbt_test")
            test = test_task(
                project_path=dbt_project_path_full,
                command=f"test {test_select_safe} {dbt_target_option}",
                raise_on_failure=False,
                wait_for=[run],
            )
            upload_metadata_upstream_task = test

    else:
        # Produce `catalog.json` and `manifest.json` artifacts for Luma ingestion.
        docs_generate_task = dbt_task.with_options(name="dbt_docs_generate")
        docs = docs_generate_task(
            project_path=dbt_project_path_full,
            command="docs generate",
            wait_for=[pull_dbt_deps],
        )
        upload_metadata_upstream_task = docs

    # Upload metadata to Luma.
    if dbt_target_dir_path is None:
        dbt_target_dir_path = dbt_project_path_full / "target"

    try:
        upload_metadata = luma_ingest_task(
            metadata_kind=metadata_kind,
            metadata_dir_path=dbt_target_dir_path,
            luma_url=luma_url,
            follow=luma_follow,
            wait_for=[upload_metadata_upstream_task],
        )
    except Exception:
        msg = "Luma ingest task failed."
        logger.exception(msg)
        upload_metadata = msg
        task_failed = True

    if run_results_storage_path:
        # Set the file path to include date info.
        file_name = "run_results.json"
        now = datetime.now(timezone.utc)
        run_results_storage_path = run_results_storage_path.rstrip("/") + "/"

        # Add partitioning.
        date_str = now.strftime("%Y%m%d")
        run_results_storage_path += date_str + "/"

        # Add timestamp suffix, eg. run_results_1737556947.934292.json.
        timestamp = now.timestamp()
        run_results_storage_path += (
            Path(file_name).stem + "_" + str(timestamp) + ".json"
        )
        logger.info(f"Uploading run results to {run_results_storage_path}")
        # Upload the file to s3.
        try:
            dump_test_results_to_s3 = s3_upload_file(
                from_path=str(dbt_target_dir_path / file_name),
                to_path=run_results_storage_path,
                wait_for=[upload_metadata_upstream_task],
                config_key=run_results_storage_config_key,
                credentials_secret=run_results_storage_credentials_secret,
            )
        except Exception:
            msg = "S3 upload file task failed."
            logger.exception(msg)
            dump_test_results_to_s3 = msg
            task_failed = True

    # Cleanup.
    wait_for = (
        [upload_metadata, dump_test_results_to_s3]
        if run_results_storage_path
        else [upload_metadata]
    )
    remove_dbt_repo_dir(dbt_repo_name, wait_for=wait_for)

    if task_failed:
        return Failed()

    if fail_flow_only_on_build_failure and build_select:
        model_error_pattern = re.compile(r"ERROR creating", re.IGNORECASE)
        if any(model_error_pattern.search(line) for line in build):
            return Failed(message="One or more models failed to build.")

    return remove_dbt_repo_dir



"""Test flows `test_eurostat_to_adls`."""

from viadot.orchestration.prefect.flows import (
    eurostat_to_adls,
)
from viadot.sources import AzureDataLake


TEST_FILE_PATH = "raw/viadot_2_0_TEST_eurostat.parquet"


def test_eurostat_to_adls():
    """Function for testing uploading data from Eurostat to ADLS."""
    lake = AzureDataLake(config_key="adls_test")

    assert not lake.exists(TEST_FILE_PATH)

    eurostat_to_adls(
        dataset_code="ILC_DI04",
        adls_path=TEST_FILE_PATH,
        adls_credentials_secret="sp-adls-test",  # noqa: S106, # pragma: allowlist secret
    )

    assert lake.exists(TEST_FILE_PATH)

    lake.rm(TEST_FILE_PATH)



"""Task for downloading data from Genesys Cloud API."""

from typing import Any

import pandas as pd
from prefect import get_run_logger, task

from viadot.exceptions import APIError
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Genesys


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=2 * 60 * 60)
def genesys_to_df(  # noqa: PLR0913
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    verbose: bool | None = None,
    endpoint: str | None = None,
    environment: str = "mypurecloud.de",
    queues_ids: list[str] | None = None,
    view_type: str | None = None,
    view_type_time_sleep: int | None = None,
    post_data_list: list[dict[str, Any]] | None = None,
    time_between_api_call: float = 0.5,
    normalization_sep: str = ".",
    drop_duplicates: bool = False,
    validate_df_dict: dict[str, Any] | None = None,
) -> pd.DataFrame:
    """Task to download data from Genesys Cloud API.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        verbose (bool, optional): Increase the details of the logs printed on the
                screen. Defaults to False.
        endpoint (Optional[str], optional): Final end point to the API.
            Defaults to None.
        environment (str, optional): the domain that appears for Genesys Cloud
            Environment based on the location of your Genesys Cloud organization.
            Defaults to "mypurecloud.de".
        queues_ids (Optional[List[str]], optional): List of queues ids to consult the
                members. Defaults to None.
        view_type (Optional[str], optional): The type of view export job to be created.
            Defaults to None.
        view_type_time_sleep (Optional[int], optional): Waiting time to retrieve data
            from Genesys Cloud API. Defaults to None.
        post_data_list (Optional[List[Dict[str, Any]]], optional): List of string
            templates to generate json body in POST calls to the API. Defaults to None.
        time_between_api_call (int, optional): The time, in seconds, to sleep the call
            to the API. Defaults to 0.5.
        normalization_sep (str, optional): Nested records will generate names separated
            by sep. Defaults to ".".
        drop_duplicates (bool, optional): Remove duplicates from the DataFrame.
            Defaults to False.
        validate_df_dict (Optional[Dict[str, Any]], optional): A dictionary with
            optional list of tests to verify the output dataframe. Defaults to None.

    Examples:
        data_frame = genesys_to_df(
            config_key=config_key,
            azure_key_vault_secret=azure_key_vault_secret,
            verbose=verbose,
            endpoint=endpoint,
            environment=environment,
            queues_ids=queues_ids,
            view_type=view_type,
            view_type_time_sleep=view_type_time_sleep,
            post_data_list=post_data_list,
            normalization_sep=normalization_sep,
            validate_df_dict=validate_df_dict,
        )

    Returns:
        pd.DataFrame: The response data as a pandas DataFrame.
    """
    logger = get_run_logger()

    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    if endpoint is None:
        msg = "The API endpoint parameter was not defined."
        raise APIError(msg)

    genesys = Genesys(
        credentials=credentials,
        config_key=config_key,
        verbose=verbose,
        environment=environment,
    )
    logger.info("running `api_connection` method:\n")
    genesys.api_connection(
        endpoint=endpoint,
        queues_ids=queues_ids,
        view_type=view_type,
        view_type_time_sleep=view_type_time_sleep,
        post_data_list=post_data_list,
        time_between_api_call=time_between_api_call,
        normalization_sep=normalization_sep,
    )
    logger.info("running `to_df` method:\n")

    return genesys.to_df(
        drop_duplicates=drop_duplicates,
        validate_df_dict=validate_df_dict,
    )



"""Test task `eurostat_to_df`."""

import pandas as pd
from prefect import flow

from viadot.orchestration.prefect.tasks import eurostat_to_df


def test_task_connection():
    """Function for testing eurstat connection."""

    @flow
    def test_eurostat_to_df():
        """Function for testing eurostat_to_df task."""
        df = eurostat_to_df(
            dataset_code="ILC_DI04",
            params={"hhtyp": "total", "indic_il": "med_e"},
        )
        assert isinstance(df, pd.DataFrame)
        assert not df.empty
        assert list(df.columns) == [
            "geo",
            "time",
            "indicator",
            "label",
            "updated",
            "_viadot_source",
            "_viadot_downloaded_at_utc",
        ]

    test_eurostat_to_df()



"""Source imports."""

from importlib.util import find_spec

from ._duckdb import DuckDB
from ._trino import Trino
from .azure_sql import AzureSQL
from .bigquery import BigQuery
from .business_core import BusinessCore
from .cloud_for_customers import CloudForCustomers
from .customer_gauge import CustomerGauge
from .ecb import ECBExchangeRates
from .entraid import EntraID
from .epicor import Epicor
from .eurostat import Eurostat
from .exchange_rates import ExchangeRates
from .genesys import Genesys
from .hubspot import Hubspot
from .matomo import Matomo
from .mediatool import Mediatool
from .mindful import Mindful
from .onestream import OneStream
from .outlook import Outlook
from .salesforce import Salesforce
from .sftp import Sftp
from .sharepoint import Sharepoint, SharepointList
from .smb import SMB
from .smb_client import SMBClient
from .sql_server import SQLServer
from .sqlite import SQLite
from .supermetrics import Supermetrics
from .tm1 import TM1
from .uk_carbon_intensity import UKCarbonIntensity
from .vid_club import VidClub


__all__ = [
    "SMB",
    "SMBClient",
    "TM1",
    "AzureSQL",
    "BigQuery",
    "BusinessCore",
    "CloudForCustomers",
    "CustomerGauge",
    "DuckDB",
    "ECBExchangeRates",
    "EntraID",
    "Epicor",
    "Eurostat",
    "ExchangeRates",
    "Genesys",
    "Hubspot",
    "Matomo",
    "Mediatool",
    "Mindful",
    "OneStream",
    "Outlook",
    "SQLServer",
    "SQLite",
    "Salesforce",
    "Sftp",
    "Sharepoint",
    "SharepointList",
    "Supermetrics",
    "Trino",
    "UKCarbonIntensity",
    "VidClub",
]

if find_spec("adlfs"):
    from viadot.sources.azure_data_lake import AzureDataLake  # noqa: F401

    __all__.extend(["AzureDataLake"])

if find_spec("redshift_connector"):
    from viadot.sources.redshift_spectrum import RedshiftSpectrum  # noqa: F401

    __all__.extend(["RedshiftSpectrum"])

if find_spec("s3fs"):
    from viadot.sources._minio import MinIO  # noqa: F401
    from viadot.sources.s3 import S3  # noqa: F401

    __all__.extend(["S3", "MinIO"])

if find_spec("pyrfc"):
    from viadot.sources.sap_bw import SAPBW  # noqa: F401
    from viadot.sources.sap_rfc import SAPRFC  # noqa: F401

    __all__.extend(["SAPBW", "SAPRFC"])

if find_spec("pyspark"):
    from viadot.sources.databricks import Databricks  # noqa: F401

    __all__.append("Databricks")



"""'customer_gauge_to_adls.py'."""

from datetime import datetime
from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import customer_gauge_to_df, df_to_adls


@flow
def customer_gauge_to_adls(  # noqa: PLR0913
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    endpoint: Literal["responses", "non-responses"] = "non-responses",
    cursor: int | None = None,
    pagesize: int = 1000,
    date_field: str | None = None,
    start_date: datetime | None = None,
    end_date: datetime | None = None,
    total_load: bool = True,
    unpack_by_field_reference_cols: list[str] | None = None,
    unpack_by_nested_dict_transformer: list[str] | None = None,
    validate_df_dict: dict[str, Any] | None = None,
    anonymize: bool = False,
    columns_to_anonymize: list[str] | None = None,
    anonymize_method: Literal["mask", "hash"] = "mask",
    anonymize_value: str = "***",
    date_column: str | None = None,
    days: int | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_config_key: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Download data from the Customer Gauge API using Prefect.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            containing a dictionary with ['client_id', 'client_secret'].
            Defaults to None.
        endpoint (Literal["responses", "non-responses"], optional): Indicate which
            endpoint to connect. Defaults to "non-responses.
        cursor (int, optional): Cursor value to navigate to the page.
            Defaults to None.
        pagesize (int, optional): Number of responses (records) returned per page,
            max value = 1000. Defaults to 1000. Defaults to 1000.
        date_field (str, optional): Specifies the date type which filter date range.
            Possible options: "date_creation", "date_order", "date_sent" or
            "date_survey_response". Defaults to None.
        start_date (datetime, optional): Defines the period start date in
            yyyy-mm-dd format. Defaults to None.
        end_date (datetime, optional): Defines the period end date in
            yyyy-mm-dd format. Defaults to None.
        total_load (bool, optional): Indicate whether to download the data to the
            latest. If 'False', only one API call is executed (up to 1000 records).
            Defaults to True.
        unpack_by_field_reference_cols (list[str]): Columns to unpack and modify using
            `_field_reference_unpacker`. Defaults to None.
        unpack_by_nested_dict_transformer (list[str]): Columns to unpack and modify
            using `_nested_dict_transformer`. Defaults to None.
        validate_df_dict (dict[str], optional): A dictionary with optional list of
            tests to verify the output dataframe. If defined, triggers the
            `validate_df` task from task_utils. Defaults to None.
        anonymize (bool, optional): Indicates if anonymize selected columns.
            Defaults to False.
        columns_to_anonymize (list[str], optional): List of columns to anonymize.
            Defaults to None.
        anonymize_method  (Literal["mask", "hash"], optional): Method of
            anonymizing data. "mask" -> replace the data with "value" arg. "hash" ->
            replace the data with the hash value of an object (using `hash()`
            method). Defaults to "mask".
        anonymize_value (str, optional): Value to replace the data.
            Defaults to "***".
        date_column (str, optional): Name of the date column used to identify rows
            that are older than a specified number of days. Defaults to None.
        days (int, optional): The number of days beyond which we want to anonymize
            the data, e.g. older than 2 years can be: 2*365. Defaults to None.
        adls_azure_key_vault_secret (str, optional): The name of the Azure Key.
            Defaults to None.
        adls_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        adls_path (str, optional): Azure Data Lake destination folder/catalog path.
            Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to False.
    """
    data_frame = customer_gauge_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        endpoint=endpoint,
        cursor=cursor,
        pagesize=pagesize,
        date_field=date_field,
        start_date=start_date,
        end_date=end_date,
        total_load=total_load,
        unpack_by_field_reference_cols=unpack_by_field_reference_cols,
        unpack_by_nested_dict_transformer=unpack_by_nested_dict_transformer,
        validate_df_dict=validate_df_dict,
        anonymize=anonymize,
        columns_to_anonymize=columns_to_anonymize,
        anonymize_method=anonymize_method,
        anonymize_value=anonymize_value,
        date_column=date_column,
        days=days,
    )

    return df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""UK Carbon Intensity connector."""

import pandas as pd
import requests

from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class UKCarbonIntensity(Source):
    def __init__(self, *args, api_url: str | None = None, **kwargs):
        """Fetch data of Carbon Intensity of the UK Power Grid.

        Documentation for this source API is located
        at: https://carbon-intensity.github.io/api-definitions/#carbon-intensity-api-v2-0-0

        Parameters
        ----------
        api_url : str, optional
        The URL endpoint to call, by default None
        """
        super().__init__(*args, **kwargs)
        self.api_url = api_url
        self.API_ENDPOINT = "https://api.carbonintensity.org.uk"

    def to_json(self) -> dict:
        """Creates json file."""
        url = f"{self.API_ENDPOINT}{self.api_url}"
        headers = {"Accept": "application/json"}
        response = requests.get(url, params={}, headers=headers, timeout=10)
        if response.ok:
            return response.json()
        raise f"Error {response.json()}"

    @add_viadot_metadata_columns
    def to_df(self, if_empty: str = "warn") -> pd.DataFrame:
        """Returns a pandas DataFrame with flattened data.

        Returns:
            pandas.DataFrame: A Pandas DataFrame
        """
        from_ = []
        to = []
        forecast = []
        actual = []
        max_ = []
        average = []
        min_ = []
        index = []
        json_data = self.to_json()

        if not json_data:
            self._handle_if_empty(if_empty)

        for row in json_data["data"]:
            from_.append(row["from"])
            to.append(row["to"])
            index.append(row["intensity"]["index"])
            try:
                forecast.append(row["intensity"]["forecast"])
                actual.append(row["intensity"]["actual"])
                df = pd.DataFrame(
                    {
                        "from": from_,
                        "to": to,
                        "forecast": forecast,
                        "actual": actual,
                        "index": index,
                    }
                )
            except KeyError:
                max_.append(row["intensity"]["max"])
                average.append(row["intensity"]["average"])
                min_.append(row["intensity"]["min"])
                df = pd.DataFrame(
                    {
                        "from": from_,
                        "to": to,
                        "max": max_,
                        "average": average,
                        "min": min_,
                    }
                )
        return df

    def query(self) -> None:
        """Queries the API."""
        ...



import os
from pathlib import Path

import pandas as pd
from prefect import flow
import pytest

from viadot.orchestration.prefect.tasks import s3_upload_file
from viadot.utils import skip_test_on_missing_extra


try:
    from viadot.sources import S3
except ImportError:
    skip_test_on_missing_extra(source_name="S3", extra="aws")

S3_BUCKET = os.environ.get("S3_BUCKET")
TEST_SCHEMA = "raw_test"
TEST_TABLE = "test_s3_upload_file"


@pytest.fixture(scope="session")
def s3(aws_config_key):
    return S3(config_key=aws_config_key)


@pytest.fixture
def TEST_FILE_PATH():
    path = "test.csv"
    df = pd.DataFrame(
        [
            [0, "A"],
            [1, "B"],
            [2, "C"],
        ],
        columns=["col1", "col2"],
    )
    df.to_csv(path)

    yield path

    Path(path).unlink()


def test_s3_upload_file(TEST_FILE_PATH):
    @flow
    def test_flow():
        file_path = f"s3://{S3_BUCKET}/nesso/{TEST_SCHEMA}/{TEST_TABLE}.csv"
        s3_upload_file(
            from_path=TEST_FILE_PATH,
            to_path=file_path,
        )

        file_exists = s3.exists(path=file_path)

        s3.rm(paths=[file_path])

        return file_exists

    assert test_flow() is True



"""Tests for the EntraID source."""

import asyncio
from unittest.mock import MagicMock

import aiohttp
import pandas as pd
import pytest

from viadot.sources.entraid import EntraID


def _base_credentials():
    return {
        "tenant_id": "tenant",
        "client_id": "client",
        "client_secret": "secret",  # pragma: allowlist secret
        "scope": "https://graph.microsoft.com/.default",
    }


class _ReqInfo:
    def __init__(self, url: str = "https://example.com"):
        """Store request URL for fake responses."""
        self.real_url = url


class MockResponse:
    def __init__(self, status=200, json_data=None, headers=None):
        """Create a fake response with status, JSON and headers."""
        self.status = status
        self._json_data = json_data or {}
        self.headers = headers or {}

    async def json(self):
        return self._json_data

    def raise_for_status(self):
        if self.status >= 400:
            raise aiohttp.ClientResponseError(
                request_info=_ReqInfo(),
                history=(),
                status=self.status,
                message="error",
                headers=self.headers,
            )


class MockCtxMgr:
    def __init__(self, response: MockResponse):
        """Wrap a response in an async context manager."""
        self._response = response

    async def __aenter__(self):
        """Enter async context and return the mock response."""
        return self._response

    async def __aexit__(self, exc_type, exc, tb):
        """Exit async context without suppressing exceptions."""
        return False


class MockSession:
    def __init__(self, get_responses=None, post_response=None):
        """Fake session yielding pre-set GET/POST responses."""
        self.headers = {}
        self._get_iter = iter(get_responses or [])
        self._post_response = post_response

    def get(self, _url, **_kwargs):
        return MockCtxMgr(next(self._get_iter))

    def post(self, _url, **_kwargs):
        return MockCtxMgr(self._post_response)


@pytest.mark.asyncio
async def test_get_access_token_success():
    """Return token from the token endpoint."""
    e = EntraID(credentials=_base_credentials())
    session = MockSession(
        post_response=MockResponse(status=200, json_data={"access_token": "tok"})
    )
    result = await e._get_access_token(session)  # type: ignore[arg-type]
    assert result == "tok"


@pytest.mark.asyncio
async def test_authorize_session_sets_header():
    """Set Authorization header after token retrieval."""
    e = EntraID(credentials=_base_credentials())
    session = MockSession(
        post_response=MockResponse(status=200, json_data={"access_token": "tok"})
    )
    await e._authorize_session(session)  # type: ignore[arg-type]
    assert session.headers.get("Authorization") == "Bearer tok"


@pytest.mark.asyncio
async def test_fetch_all_pages_pagination_with_retry():
    """Retry on 429 and follow @odata.nextLink."""
    e = EntraID(credentials=_base_credentials())
    # 1st call -> 429 with Retry-After=0, then 2 pages 200
    r1 = MockResponse(status=429, headers={"Retry-After": "0"})
    r2 = MockResponse(
        status=200, json_data={"value": [{"a": 1}], "@odata.nextLink": "next"}
    )
    r3 = MockResponse(status=200, json_data={"value": [{"a": 2}]})
    session = MockSession(get_responses=[r1, r2, r3])
    items = await e._fetch_all_pages(session, "http://test")  # type: ignore[arg-type]
    assert items == [{"a": 1}, {"a": 2}]


@pytest.mark.asyncio
async def test_get_all_users_calls_fetch_with_url(monkeypatch):
    """Call users endpoint with expected select fields."""
    e = EntraID(credentials=_base_credentials())
    captured = {}

    async def fake_fetch(*args, **kwargs):
        # Method may be called with (session, url) or url as kw
        if len(args) >= 2:
            captured["url"] = args[1]
        else:
            captured["url"] = kwargs["url"]
        return [{"id": "1", "mail": "a@b.com"}]

    monkeypatch.setattr(e, "_fetch_all_pages", fake_fetch)  # type: ignore[attr-defined]

    session = MagicMock()
    res = await e._get_all_users(session)  # type: ignore[arg-type]
    assert res == [{"id": "1", "mail": "a@b.com"}]
    assert captured["url"] == "https://graph.microsoft.com/v1.0/users?$select=id,mail"


@pytest.mark.asyncio
async def test_get_user_group_membership_maps_fields(monkeypatch):
    """Map group entries to user_email and group_name fields."""
    e = EntraID(credentials=_base_credentials())

    async def fake_fetch(*_args, **_kwargs):
        return [{"displayName": "G1"}, {"displayName": "G2"}]

    monkeypatch.setattr(e, "_fetch_all_pages", fake_fetch)  # type: ignore[attr-defined]
    session = MagicMock()
    user = {"id": "u1", "mail": "x@y.com"}
    res = await e._get_user_group_membership(session, user)  # type: ignore[arg-type]
    assert res == [
        {"user_email": "x@y.com", "group_name": "G1"},
        {"user_email": "x@y.com", "group_name": "G2"},
    ]


@pytest.mark.asyncio
async def test_concat_user_groups_df_aggregates(monkeypatch):
    """Aggregate group rows for provided users."""
    e = EntraID(credentials=_base_credentials())

    async def fake_get_groups(_session, user):
        if user["id"] == "1":
            return [{"user_email": user["mail"], "group_name": "A"}]
        return [{"user_email": user["mail"], "group_name": "B"}]

    monkeypatch.setattr(e, "_get_user_group_membership", fake_get_groups)  # type: ignore[attr-defined]

    users_df = pd.DataFrame(
        [{"id": "1", "mail": "a@b.com"}, {"id": "2", "mail": "c@d.com"}]
    )
    session = MagicMock()
    df = await e._concat_user_groups_df(
        session, users_df, ["user_email", "group_name"], max_concurrent_requests=5
    )  # type: ignore[arg-type]
    assert list(df.columns) == ["user_email", "group_name"]
    assert set(df["group_name"]) == {"A", "B"}


@pytest.mark.asyncio
async def test_build_users_groups_df_composes(monkeypatch):
    """Compose authorize, get users, and build groups steps."""
    e = EntraID(credentials=_base_credentials())

    async def fake_authorize(*_args, **_kwargs):
        return None

    async def fake_get_users(*_args, **_kwargs):
        return [{"id": "1", "mail": "a@b.com"}]

    async def fake_build(*_args, **_kwargs):
        return pd.DataFrame([{"user_email": "a@b.com", "group_name": "A"}])

    monkeypatch.setattr(e, "_authorize_session", fake_authorize)  # type: ignore[attr-defined]
    monkeypatch.setattr(e, "_get_all_users", fake_get_users)  # type: ignore[attr-defined]
    monkeypatch.setattr(e, "_concat_user_groups_df", fake_build)  # type: ignore[attr-defined]

    df = await e._build_users_groups_df(max_concurrent_requests=3)
    assert not df.empty
    assert set(df.columns) == {"user_email", "group_name"}


def test_run_fallback_when_event_loop_running(monkeypatch):
    """Use fallback loop when an event loop is already running."""
    e = EntraID(credentials=_base_credentials())

    async def sample_coro():
        return 123

    class FakeLoop:
        def __init__(self):
            self.closed = False

        def run_until_complete(self, _coro):
            return 123

        def close(self):
            self.closed = True

    def fake_new_loop():
        return FakeLoop()

    def fake_set_loop(_loop):
        return None

    def fake_asyncio_run(_coro):
        msg = "asyncio.run() cannot be called from a running event loop"
        raise RuntimeError(msg)

    monkeypatch.setattr(asyncio, "run", fake_asyncio_run)
    monkeypatch.setattr(asyncio, "new_event_loop", fake_new_loop)
    monkeypatch.setattr(asyncio, "set_event_loop", fake_set_loop)

    result = e._run(sample_coro())
    assert result == 123


def test_to_df_calls_cleanup_and_validate(monkeypatch):
    """Call remove_newlines_and_tabs and validate during to_df."""
    e = EntraID(credentials=_base_credentials())

    # Return a small DataFrame from the async builder
    async def fake_build(**_kwargs):
        return pd.DataFrame([{"user_email": "a@b.com", "group_name": "A"}])

    monkeypatch.setattr(e, "_build_users_groups_df", fake_build)  # type: ignore[attr-defined]

    called = {"cleanup": False, "validate": False}

    def fake_cleanup(df):
        called["cleanup"] = True
        return df

    def fake_validate(df, tests):  # noqa: ARG001
        called["validate"] = True
        assert tests == {"k": "v"}

    # Patch symbols as imported in module namespace
    import viadot.sources.entraid as ent_mod

    monkeypatch.setattr(ent_mod, "remove_newlines_and_tabs", fake_cleanup)
    monkeypatch.setattr(ent_mod, "validate", fake_validate)

    out = e.to_df(if_empty="warn", tests={"k": "v"}, max_concurrent_requests=2)
    assert not out.empty
    assert called["cleanup"] is True
    assert called["validate"] is True



from viadot.orchestration.prefect.flows import sql_server_to_minio
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import MinIO


PATH = "data/tables_schemas.parquet"


def test_sql_server_to_minio():
    credentials = get_credentials("minio-dev")
    sql_server_to_minio(
        query=""" SELECT t.name as table_name
                ,s.name as schema_name
                FROM sys.tables t
                JOIN sys.schemas s
                ON t.schema_id = s.schema_id""",
        path="s3://datalake-dev/data/tables_schemas.parquet",
        sql_server_credentials_secret="sql-server",  # noqa: S106
        minio_credentials=credentials,
        basename_template="test-{i}",
        if_exists="overwrite_or_ignore",
    )

    minio = MinIO(credentials=credentials)
    file_exists = minio._check_if_file_exists(PATH)

    assert file_exists is True

    minio.rm(path=PATH)



"""Source for connecting to Epicor Prelude API."""

from typing import Any, Literal
from xml.etree.ElementTree import Element as ElementT

from defusedxml import ElementTree as ElementDE
import pandas as pd
from pydantic import BaseModel
import requests

from viadot.config import get_source_credentials
from viadot.exceptions import DataRangeError, ValidationError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_response


def parse_orders_xml(response: requests.models.Response) -> pd.DataFrame:
    """Parse XML order data into a pandas DataFrame.

    Args:
        response (requests.models.Response): Response from Epicor API.

    Returns:
        pd.DataFrame: xml data in a form of pandas DataFrame.
    """
    root = ElementDE.fromstring(response.text)
    rows = []

    for child in root:
        header = child.find("HeaderInformation")
        header_data = flatten_element(header, prefix="HeaderInformation.")

        line_items_parent = child.find("LineItemDetails")
        line_items = (
            line_items_parent.findall("LineItemDetail")
            if line_items_parent is not None
            else []
        )

        if not line_items:
            rows.append(header_data)
            continue

        for item in line_items:
            item_data = flatten_element(item, prefix="LineItemDetail.")
            row = {**header_data, **item_data}
            rows.append(row)

    return pd.DataFrame(rows)


def parse_customer_xml(response: requests.models.Response) -> pd.DataFrame:
    """Parse XML customer data into a pandas DataFrame.

    Args:
        response (requests.models.Response): Response from Epicor API.

    Returns:
        pd.DataFrame: xml data in a form of pandas DataFrame.
    """
    xml_str = response.text
    root = ElementDE.fromstring(xml_str)

    rows = []

    for cust in root.findall(".//Customer"):
        customer_fields = {}
        for child in cust:
            if child.tag != "ShipTos" and len(child) == 0:
                customer_fields[child.tag] = child.text

        shiptos_parent = cust.find("ShipTos")
        shiptos = (
            shiptos_parent.findall("ShipToAddress")
            if shiptos_parent is not None
            else []
        )

        if shiptos:
            for st in shiptos:
                row = customer_fields.copy()

                # flatten shipto fields
                for elem in st:
                    row[f"ShipToAddress.{elem.tag}"] = elem.text

                rows.append(row)

        else:
            # no shiptos: return only customer flat row
            rows.append(customer_fields)

    return pd.DataFrame(rows)


def flatten_element(elem: ElementT, prefix: str = "") -> dict[str, str]:
    """Flatten element tree into a dict.

    Args:
        elem (str): XML element to flatten.
        prefix: Prefix for building dotted tag paths.

    Returns:
        dict[str, str]: Mapping of dotted tag paths to leaf text values.
    """
    if elem is None:
        return {}

    data = {}
    for child in elem:
        tag = f"{prefix}{child.tag}"

        if len(child):
            nested = flatten_element(child, prefix=tag + ".")
            data.update(nested)
        else:
            data[tag] = child.text

    return data


class EpicorCredentials(BaseModel):
    host: str
    port: int = 443
    username: str
    password: str


class Epicor(Source):
    """Source for connecting to Epicor Prelude API."""

    def __init__(
        self,
        base_url: str,
        credentials: dict[str, Any] | None = None,
        config_key: str | None = None,
        validate_date_filter: bool = True,
        start_date_field: str = "BegInvoiceDate",
        end_date_field: str = "EndInvoiceDate",
        *args,
        **kwargs,
    ):
        """Class to connect to Epicor API and pasere XML output into a pandas DataFrame.

        Args:
            base_url (str, required): Base url to Epicor.
            filters_xml (str, required): Filters in form of XML. The date filter
                 is required.
            credentials (dict[str, Any], optional): Credentials to connect with
            Epicor API containing host, port, username and password.Defaults to None.
            config_key (str, optional): Credential key to dictionary where details
                 are stored.
            validate_date_filter (bool, optional): Whether or not validate xml
                date filters. Defaults to True.
            start_date_field (str, optional) The name of filters field containing
                start date.Defaults to "BegInvoiceDate".
            end_date_field (str, optional) The name of filters field containing
                end date. Defaults to "EndInvoiceDate".
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(EpicorCredentials(**raw_creds))

        self.base_url = base_url
        self.validate_date_filter = validate_date_filter
        self.start_date_field = start_date_field
        self.end_date_field = end_date_field

        self.url = (
            "http://"
            + validated_creds["host"]
            + ":"
            + str(validated_creds["port"])
            + base_url
        )

        super().__init__(*args, credentials=validated_creds, **kwargs)

    def generate_token(self) -> str:
        """Function to generate API access token that is valid for 24 hours.

        Returns:
            str: Generated token.
        """
        url = (
            "http://"
            + self.credentials["host"]
            + ":"
            + str(self.credentials["port"])
            + "/api/security/token/"
        )

        headers = {
            "Content-Type": "application/json",
            "username": self.credentials["username"],
            "password": self.credentials["password"],
        }

        response = handle_api_response(url=url, headers=headers, method="POST")
        root = ElementDE.fromstring(response.text)
        return root.find("AccessToken").text

    def validate_filter(self, filters_xml: str) -> None:
        """Validate that the XML filter contains required date fields.

        Args:
            filters_xml: XML string with filter parameters.

        Raises:
            DataRangeError: If start or end date is missing.

        """
        root = ElementDE.fromstring(filters_xml)
        for child in root:
            for subchild in child:
                if (
                    subchild.tag in (self.start_date_field, self.end_date_field)
                ) and subchild.text is None:
                    msg = "Too much data. Please provide a date range filter."
                    raise DataRangeError(msg)

    def get_xml_response(self, filters_xml: str) -> requests.models.Response:
        """Send the XML filter request to the Epicor API and return the response.

        Args:
            filters_xml: XML string with filter parameters.

        Returns:
            Response: HTTP response from the Epicor API.
        """
        if self.validate_date_filter is True:
            self.validate_filter(filters_xml)
        payload = filters_xml
        headers = {
            "Content-Type": "application/xml",
            "Authorization": "Bearer " + self.generate_token(),
        }
        return handle_api_response(
            url=self.url, headers=headers, data=payload, method="POST"
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        filters_xml: str,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
    ) -> pd.DataFrame:
        """Convert Epicor API XML response to pandas DataFrame.

        Args:
            filters_xml (str): XML with query filters sent to the API.
            if_empty (Literal["warn", "skip", "fail"], optional):
                Behavior when the parsed DataFrame is empty. Defaults to "warn".

        Returns:
            pd.DataFrame: Parsed data returned by the appropriate XML parser.

        Raises:
            ValidationError: If no parser is available for the given view.

        """
        data = self.get_xml_response(filters_xml)

        if (
            "ORDER.HISTORY.DETAIL.QUERY" in self.base_url
            or "ORDER.DETAIL.PROD.QUERY" in self.base_url
            or "BOOKINGS.DETAIL.QUERY" in self.base_url
        ):
            df = parse_orders_xml(data)
        elif "CUSTOMER.QUERY" in self.base_url:
            df = parse_customer_xml(data)
        else:
            msg = f"Parser for selected view {self.base_url} is not available"
            raise ValidationError(msg)
        if df.empty:
            self._handle_if_empty(if_empty)
        return df



"""Flow for extracting data from the DuckDB into SQLServer."""

from pathlib import Path
from typing import Any, Literal

from prefect import flow, task
from prefect.logging import get_run_logger

from viadot.orchestration.prefect.tasks import (
    bcp,
    create_sql_server_table,
    duckdb_query,
)
from viadot.orchestration.prefect.tasks.task_utils import (
    df_to_csv,
    get_sql_dtypes_from_df,
)


@task(timeout_seconds=60 * 60)
def cleanup_csv_task(path: str) -> None:
    """Remove a CSV file from the local filesystem."""
    logger = get_run_logger()

    logger.info(f"Removing file {path}...")
    try:
        Path(path).unlink()
        logger.info(f"File {path} has been successfully removed.")
    except Exception:
        logger.exception(f"File {path} could not be removed.")


@flow(
    name="extract--duckdb--sql_server",
    description="Extract data from DuckDB and save it in the SQLServer",
    retries=1,
    retry_delay_seconds=60,
    timeout_seconds=2 * 60 * 60,
)
def duckdb_to_sql_server(  # noqa: PLR0913
    query: str,
    local_path: str,
    db_table: str,
    db_schema: str,
    if_exists: Literal["fail", "replace", "skip", "delete"] = "replace",
    dtypes: dict[str, Any] | None = None,
    chunksize: int = 5000,
    error_log_file_path: str = "./log_file.log",
    on_error: Literal["skip", "fail"] = "skip",
    duckdb_credentials_secret: str | None = None,
    # Specifying credentials in a dictionary is not recommended in the viadot flows,
    # but in this case credentials can include only database name.
    duckdb_credentials: dict[str, Any] | None = None,
    duckdb_config_key: str | None = None,
    sql_server_credentials_secret: str | None = None,
    sql_server_config_key: str | None = None,
) -> None:
    """Download a table from DuckDB and upload it to the SQLServer.

    Args:
        query (str, required): The query to execute on the SQL Server database.
            If the qery doesn't start with "SELECT" returns an empty DataFrame.
        local_path (str): Where to store the CSV data dump used for bulk upload to SQL
            Server.
        db_table (str, optional): Destination table. Defaults to None.
        db_schema (str, optional): Destination schema. Defaults to None.
        if_exists (Literal, optional): What to do if the table exists. Defaults to
            "replace".
        dtypes (dict, optional): The data types to be enforced for the resulting table.
            By default, inferred from the DataFrame. Defaults to None.
        chunksize (int, optional): Size of a chunk to use in the bcp function.
            Defaults to 5000.
        error_log_file_path (string, optional): Full path of an error file. Defaults
            to "./log_file.log".
        on_error (str, optional): What to do in case of a bcp error. Defaults to "skip".
        duckdb_credentials_secret (str, optional): The name of the secret storing
            the credentials to the DuckDB. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        duckdb_credentials (dict[str, Any], optional): Credentials to the DuckDB.
            Defaults to None.
        duckdb_config_key (str, optional): The key in the viadot config holding relevant
            credentials to the DuckDB. Defaults to None.
        sql_server_credentials_secret (str, optional): The name of the secret storing
            the credentials to the SQLServer. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        sql_server_config_key (str, optional): The key in the viadot config holding
            relevant credentials to the SQLServer. Defaults to None.

    """
    df = duckdb_query(
        query=query,
        fetch_type="dataframe",
        credentials=duckdb_credentials,
        config_key=duckdb_config_key,
        credentials_secret=duckdb_credentials_secret,
    )
    if dtypes is None:
        dtypes = get_sql_dtypes_from_df(df)

    create_sql_server_table(
        table=db_table,
        schema=db_schema,
        dtypes=dtypes,
        if_exists=if_exists,
        credentials_secret=sql_server_credentials_secret,
        config_key=sql_server_config_key,
    )
    df_to_csv(df=df, path=local_path)

    bcp(
        path=local_path,
        schema=db_schema,
        table=db_table,
        chunksize=chunksize,
        error_log_file_path=error_log_file_path,
        on_error=on_error,
        credentials_secret=sql_server_credentials_secret,
        config_key=sql_server_config_key,
    )

    cleanup_csv_task(path=local_path)



from datetime import datetime
import json
import unittest
from unittest.mock import MagicMock, patch

import pandas as pd
from pandas.testing import assert_frame_equal
import pytest
from requests.models import Response

from viadot.exceptions import APIError
from viadot.sources import Hubspot
from viadot.sources.hubspot import HubspotCredentials


variables = {
    "credentials": {"token": "fake_token"},
    "filters": [
        {
            "filters": [
                {
                    "propertyName": "createdate",
                    "operator": "GTE",
                    "value": "2021-01-01",
                }
            ]
        }
    ],
}


class TestHubspotCredentials:
    """Test Hubspot Credentials Class."""

    def test_hubspot_credentials(self):
        """Test Hubspot credentials."""
        HubspotCredentials(token="test_token")  # noqa: S106


class TestHubspot(unittest.TestCase):
    """Test Hubspot Class."""

    @classmethod
    def setUpClass(cls):  # noqa: ANN206
        """Defined based Hubspot Class for the rest of test."""
        cls.hubspot_instance = Hubspot(credentials=variables["credentials"])

    @patch("viadot.sources.hubspot.get_source_credentials", return_value=None)
    def test_init_no_credentials(self, mock_get_source_credentials):
        """Test raise error without credentials."""
        with pytest.raises(TypeError):
            Hubspot()

        mock_get_source_credentials.assert_called_once()

    def test_date_to_unix_millis(self):
        """Test Hubspot `_date_to_unix_millis` function."""
        date_str = "2021-01-01"
        expected_timestamp = int(
            datetime.strptime(date_str, "%Y-%m-%d").timestamp() * 1000
        )
        result = self.hubspot_instance._date_to_unix_millis(date_str)
        assert result == expected_timestamp

    def test_get_api_url(self):
        """Test Hubspot `_get_api_url` function."""
        endpoint = "deals"
        filters = None
        properties = ["property1", "property2"]
        expected_url = (
            f"https://api.hubapi.com/crm/v3/objects/{endpoint}/"
            + "?limit=100&properties=property1,property2&"
        )
        result = self.hubspot_instance._get_api_url(
            endpoint=endpoint, filters=filters, properties=properties
        )
        assert result == expected_url

    def test_format_filters(self):
        """Test Hubspot `_format_filters` function."""
        filters = variables["filters"]
        formatted_filters = self.hubspot_instance._format_filters(filters)
        assert isinstance(formatted_filters, list)

    def test_get_api_body(self):
        """Test Hubspot `_get_api_body` function."""
        filters = variables["filters"]
        expected_body = json.dumps({"filterGroups": filters, "limit": 100})
        result = self.hubspot_instance._get_request_body(filters)

        assert result == expected_body

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_api_call_success(self, mock_handle_api_response):
        """Test Hubspot `_api_call` method."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"results": [{"id": "123"}]}
        mock_handle_api_response.return_value = mock_response

        url = "https://api.hubapi.com/crm/v3/objects/deals/?limit=100&"
        result = self.hubspot_instance._api_call(url=url, method="GET")

        assert result == {"results": [{"id": "123"}]}
        mock_handle_api_response.assert_called_once()

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_api_call_error(self, mock_handle_api_response):
        """Test Hubspot `_api_call` method failure."""
        mock_handle_api_response.side_effect = APIError("Internal Server Error")

        url = "https://api.hubapi.com/crm/v3/objects/deals/?limit=100&"

        with pytest.raises(APIError):
            self.hubspot_instance._api_call(url=url, method="GET")

    def test_get_offset_from_response(self):
        """Test Hubspot `_get_offset_from_response` function."""
        response_with_paging = {"paging": {"next": {"after": "123"}}}
        response_with_offset = {"offset": "456"}

        offset_type, offset_value = self.hubspot_instance._get_offset_from_response(
            response_with_paging
        )
        assert (offset_type, offset_value) == ("after", "123")

        offset_type, offset_value = self.hubspot_instance._get_offset_from_response(
            response_with_offset
        )
        assert (offset_type, offset_value) == ("offset", "456")

        offset_type, offset_value = self.hubspot_instance._get_offset_from_response({})
        assert (offset_type, offset_value) == (None, None)

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_fetch_with_filters(self, mock_handle_api_response):
        """Test Hubspot `_fetch` method, with filters."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"results": [{"id": "123"}]}
        mock_handle_api_response.return_value = mock_response

        endpoint = "deals"
        filters = variables["filters"]
        properties = ["property1"]
        result = self.hubspot_instance._fetch(
            endpoint=endpoint, filters=filters, properties=properties
        )

        assert result is not None
        assert len(result) > 0

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_fetch_without_filters(self, mock_handle_api_response):
        """Test Hubspot `_fetch` method, without filters."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"results": [{"id": "123"}]}
        mock_handle_api_response.return_value = mock_response

        endpoint = "deals"
        filters = None
        properties = ["property1"]
        result = self.hubspot_instance._fetch(
            endpoint=endpoint, filters=filters, properties=properties
        )

        assert result is not None
        assert len(result) > 0

    @patch("viadot.sources.base.Source.to_df")
    def test_to_df(self, mock_parent_to_df):
        """Test Hubspot `to_df` function."""
        mock_parent_to_df.return_value = pd.DataFrame()
        result_df = self.hubspot_instance.to_df(data=[{"id": "123"}])
        result_df.drop(
            columns=["_viadot_source", "_viadot_downloaded_at_utc"],
            inplace=True,
            axis=1,
        )

        expected_df = pd.DataFrame([{"id": "123"}])
        assert_frame_equal(result_df, expected_df, check_dtype=False)
        mock_parent_to_df.assert_called_once()

    @patch("viadot.sources.base.Source.to_df")
    def test_to_df_empty(self, mock_parent_to_df):
        """Test Hubspot `to_df` method, checking emptiness."""
        mock_parent_to_df.return_value = pd.DataFrame()

        with patch.object(
            self.hubspot_instance, "_handle_if_empty"
        ) as mock_handle_if_empty:
            result_df = self.hubspot_instance.to_df(data=[])
            result_df.drop(
                columns=["_viadot_source", "_viadot_downloaded_at_utc"],
                inplace=True,
                axis=1,
            )
            mock_handle_if_empty.assert_called_once_with(
                if_empty="warn", message="The response does not contain any data."
            )
            assert result_df.empty
            mock_parent_to_df.assert_called_once()

    def test_get_api_url_full_url_passthrough(self):
        """Full URL is returned unchanged."""
        url = "https://api.hubapi.com/some/path?x=1"
        assert self.hubspot_instance._get_api_url(endpoint=url) == url

    def test_get_api_url_with_filters_builds_search_url(self):
        """Relative endpoint with filters builds search URL."""
        result = self.hubspot_instance._get_api_url(endpoint="deals", filters=[{}])
        assert result.startswith(
            "https://api.hubapi.com/crm/v3/objects/deals/search/?limit=100&"
        )

    def test_get_api_url_hubdb_prefix(self):
        """HubDB endpoints are prefixed directly."""
        endpoint = "hubdb/api/v2/tables"
        expected = f"https://api.hubapi.com/{endpoint}"
        assert self.hubspot_instance._get_api_url(endpoint=endpoint) == expected

    def test_get_api_url_missing_endpoint_raises(self):
        """Missing endpoint raises ValueError."""
        with pytest.raises(ValueError, match="Endpoint must be provided."):
            self.hubspot_instance._get_api_url()

    def test_format_filters_none_returns_empty(self):
        """None filters return empty list."""
        assert self.hubspot_instance._format_filters(None) == []

    def test_extract_items_prefers_results(self):
        """_extract_items prefers 'results' key."""
        data = {"results": [{"id": "1"}], "contacts": [{"id": "x"}]}
        items = self.hubspot_instance._extract_items(data)  # type: ignore[attr-defined]
        assert items == [{"id": "1"}]

    def test_extract_items_fallback_first_list(self):
        """_extract_items falls back to first list-valued key."""
        data = {"contacts": [{"id": "x"}], "other": "val"}
        items = self.hubspot_instance._extract_items(data)  # type: ignore[attr-defined]
        assert items == [{"id": "x"}]

    def test_build_headers_contains_token(self):
        """Authorization header includes token."""
        headers = self.hubspot_instance._build_headers()  # type: ignore[attr-defined]
        assert headers["Authorization"].endswith(variables["credentials"]["token"])
        assert headers["Content-Type"] == "application/json"

    def test_fetch_pagination_with_filters(self):
        """POST pagination with 'paging.next.after' aggregates pages."""
        instance = self.hubspot_instance
        page1 = {"results": [{"id": "1"}], "paging": {"next": {"after": "abc"}}}
        page2 = {"results": [{"id": "2"}]}
        with patch.object(instance, "_api_call", side_effect=[page1, page2]):
            result = instance._fetch(
                endpoint="deals", filters=variables["filters"], nrows=10
            )  # type: ignore[attr-defined]
        assert len(result) == 2
        assert result[0]["id"] == "1"
        assert result[1]["id"] == "2"

    def test_fetch_pagination_without_filters_offset(self):
        """GET pagination with 'offset' aggregates pages."""
        instance = self.hubspot_instance
        page1 = {"contacts": [{"id": "1"}], "offset": "2"}
        page2 = {"contacts": [{"id": "2"}]}
        with patch.object(instance, "_api_call", side_effect=[page1, page2]):
            result = instance._fetch(endpoint="contacts", filters=None, nrows=10)  # type: ignore[attr-defined]
        assert len(result) == 2
        assert result[0]["id"] == "1"
        assert result[1]["id"] == "2"

    def test_fetch_contact_ids(self):
        """Build rows with campaign_id, contact_id, contact_type."""
        instance = self.hubspot_instance
        campaigns = ["A", "B"]
        with patch.object(
            instance,
            "_api_call",
            return_value={"results": [{"id": "c1"}, {"id": "c2"}]},
        ):
            result = instance._fetch_contact_ids(campaign_ids=campaigns)  # default type
        assert len(result) == 4
        assert {r["campaign_id"] for r in result} == set(campaigns)
        assert all("contact_id" in r for r in result)
        assert all("contact_type" in r for r in result)

    def test_get_campaign_budget_totals(self):
        """Keep only totals and campaign_id."""
        instance = self.hubspot_instance
        responses = [
            {
                "budgetItems": [{"amount": 1}],
                "spendItems": [{"amount": 2}],
                "currencyCode": "USD",
                "budgetTotal": 100,
                "remainingBudget": 60,
                "spendTotal": 40,
            },
            {
                "budgetItems": [],
                "spendItems": [],
                "currencyCode": "EUR",
                "budgetTotal": 200,
                "remainingBudget": 150,
                "spendTotal": 50,
            },
        ]
        with patch.object(instance, "_api_call", side_effect=responses):
            result = instance._get_campaign_budget_totals(campaign_ids=["X", "Y"])  # type: ignore[attr-defined]
        assert len(result) == 2
        row0 = result[0]
        assert row0["campaign_id"] == "X"
        assert row0["budgetTotal"] == 100
        assert row0["remainingBudget"] == 60
        assert row0["spendTotal"] == 40
        assert row0["currencyCode"] == "USD"
        row1 = result[1]
        assert row1["campaign_id"] == "Y"
        assert row1["budgetTotal"] == 200
        assert row1["remainingBudget"] == 150
        assert row1["spendTotal"] == 50
        assert row1["currencyCode"] == "EUR"

    def test_call_api_dispatch_get_campaign_budget_totals(self):
        instance = self.hubspot_instance
        with patch.object(instance, "_get_campaign_budget_totals") as mock_gbt:
            instance.call_api(method="get_campaign_budget_totals", campaign_ids=["Z"])
            mock_gbt.assert_called_once_with(campaign_ids=["Z"])

    def test_get_campaign_metrics(self):
        """Copy selected metric keys and attach campaign_id."""
        instance = self.hubspot_instance
        with patch.object(
            instance,
            "_api_call",
            return_value={"sessions": 5, "influencedContacts": 3, "extra": "x"},
        ):
            result = instance._get_campaign_metrics(campaign_ids=["X", "Y"])
        assert len(result) == 2
        for r in result:
            assert "campaign_id" in r
            assert r["sessions"] == 5
            assert r["influencedContacts"] == 3
            assert "extra" not in r

    def test_get_campaign_details_defaults(self):
        """Fetch campaign details with default properties."""
        instance = self.hubspot_instance
        responses = [
            {
                "properties": {
                    "hs_name": "Camp A",
                    "hs_start_date": "2021-01-01",
                    "hs_end_date": "2021-02-01",
                    "hs_notes": "Note",
                    "hs_owner": "Owner",
                }
            },
            {
                "properties": {
                    "hs_name": "Camp B",
                    "hs_start_date": "2021-03-01",
                    "hs_end_date": "2021-04-01",
                    "hs_notes": "Note",
                    "hs_owner": "Owner",
                }
            },
        ]
        with patch.object(instance, "_api_call", side_effect=responses):
            result = instance._get_campaign_details(campaign_ids=["X", "Y"])  # type: ignore[attr-defined]
        assert len(result) == 2
        for row in result:
            assert "campaign_id" in row
            assert "hs_name" in row
            assert "hs_start_date" in row
            assert "hs_end_date" in row
            assert "hs_notes" in row
            assert "hs_owner" in row

    def test_get_campaign_details_empty_properties(self):
        """Handles empty 'properties' gracefully (only campaign_id present)."""
        instance = self.hubspot_instance
        responses = [{"properties": {}}, {}]
        with patch.object(instance, "_api_call", side_effect=responses):
            result = instance._get_campaign_details(campaign_ids=["X", "Y"])  # type: ignore[attr-defined]
        assert len(result) == 2
        for row, cid in zip(result, ["X", "Y"]):
            assert row["campaign_id"] == cid
            # All props keys should be present with None values
            assert "hs_name" in row
            assert row["hs_name"] is None
            assert "hs_start_date" in row
            assert row["hs_start_date"] is None
            assert "hs_end_date" in row
            assert row["hs_end_date"] is None
            assert "hs_notes" in row
            assert row["hs_notes"] is None
            assert "hs_owner" in row
            assert row["hs_owner"] is None

    def test_call_api_dispatch_get_campaign_details(self):
        instance = self.hubspot_instance
        with patch.object(instance, "_get_campaign_details") as mock_gcd:
            instance.call_api(method="get_campaign_details", campaign_ids=["Z"])
            mock_gcd.assert_called_once_with(campaign_ids=["Z"])

    def test_call_api_dispatch_get_all_contacts(self):
        instance = self.hubspot_instance
        with patch.object(instance, "_fetch") as mock_fetch:
            instance.call_api(method="get_all_contacts")
            mock_fetch.assert_called_once()
            args, kwargs = mock_fetch.call_args
            assert (
                kwargs["endpoint"]
                == "https://api.hubapi.com/contacts/v1/lists/all/contacts/all"
            )

    def test_call_api_dispatch_fetch_contact_ids(self):
        instance = self.hubspot_instance
        with patch.object(instance, "_fetch_contact_ids") as mock_fc:
            instance.call_api(
                method="fetch_contact_ids",
                campaign_ids=["Z"],
                contact_type="influencedContacts",
            )
            mock_fc.assert_called_once_with(
                campaign_ids=["Z"], contact_type="influencedContacts"
            )

    def test_call_api_dispatch_get_campaign_metrics(self):
        instance = self.hubspot_instance
        with patch.object(instance, "_get_campaign_metrics") as mock_gcm:
            instance.call_api(method="get_campaign_metrics", campaign_ids=["Z"])
            mock_gcm.assert_called_once_with(campaign_ids=["Z"])

    def test_call_api_dispatch_generic_fetch(self):
        instance = self.hubspot_instance
        with patch.object(instance, "_fetch") as mock_fetch:
            instance.call_api(
                method=None, endpoint="deals", filters=None, properties=["id"], nrows=10
            )
            mock_fetch.assert_called_once()
            args, kwargs = mock_fetch.call_args
            assert kwargs["endpoint"] == "deals"
            assert kwargs["properties"] == ["id"]
            assert kwargs["nrows"] == 10

    @patch("viadot.sources.base.Source.to_df")
    def test_to_df_includes_metadata_columns(self, mock_parent_to_df):
        mock_parent_to_df.return_value = pd.DataFrame()
        df = self.hubspot_instance.to_df(data=[{"id": "x"}])
        assert "_viadot_source" in df.columns
        assert "_viadot_downloaded_at_utc" in df.columns


if __name__ == "__main__":
    unittest.main()



from src.viadot.orchestration.prefect.flows import vid_club_to_adls
from src.viadot.sources import AzureDataLake


TEST_FILE_PATH = "test/path/to/adls.parquet"
TEST_SOURCE = "jobs"
TEST_FROM_DATE = "2023-01-01"
TEST_TO_DATE = "2023-12-31"
ADLS_CREDENTIALS_SECRET = "test_adls_secret"  # pragma: allowlist secret # noqa: S105
VIDCLUB_CREDENTIALS_SECRET = (
    "test_vidclub_secret"  # pragma: allowlist secret # noqa: S105
)


def test_vid_club_to_adls():
    lake = AzureDataLake(config_key="adls_test")

    assert not lake.exists(TEST_FILE_PATH)

    vid_club_to_adls(
        endpoint=TEST_SOURCE,
        from_date=TEST_FROM_DATE,
        to_date=TEST_TO_DATE,
        adls_path=TEST_FILE_PATH,
        adls_azure_key_vault_secret=ADLS_CREDENTIALS_SECRET,
        vidclub_credentials_secret=VIDCLUB_CREDENTIALS_SECRET,
    )

    assert lake.exists(TEST_FILE_PATH)

    lake.rm(TEST_FILE_PATH)



import pytest

from viadot.orchestration.prefect.tasks import (
    create_sql_server_table,
    sql_server_query,
    sql_server_to_df,
)
from viadot.sources import SQLServer


TABLE = "test"
SCHEMA = "sandbox"


@pytest.fixture
def sql_server():
    # Initialize the SQLServer instance with the test credentials.
    return SQLServer(config_key="sql_server")


def test_sql_server_to_df():
    df = sql_server_to_df(
        query="""
                SELECT t.name as table_name
                ,s.name as schema_name
                FROM sys.tables t
                JOIN sys.schemas s
                ON t.schema_id = s.schema_id""",
        credentials_secret="sql-server",  # noqa: S106
    )

    assert not df.empty


def test_create_sql_server_table(sql_server):
    dtypes = {
        "date": "DATE",
        "name": "VARCHAR(255)",
        "id": "VARCHAR(255)",
        "weather": "FLOAT(24)",
        "rain": "FLOAT(24)",
        "temp": "FLOAT(24)",
        "summary": "VARCHAR(255)",
    }
    create_sql_server_table(
        table=TABLE,
        schema=SCHEMA,
        dtypes=dtypes,
        if_exists="replace",
        credentials_secret="sql-server",  # noqa: S106
    )

    assert sql_server.exists(table=TABLE, schema=SCHEMA)

    sql_server_query(
        query=f"""DROP TABLE {SCHEMA}.{TABLE}""",
        credentials_secret="sql-server",  # noqa: S106
    )

    assert not sql_server.exists(table=TABLE, schema=SCHEMA)



"""Entra ID (previously Azure AD) connector."""

import asyncio
from collections.abc import Coroutine
from typing import Any, Literal, TypeVar

import aiohttp
import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    handle_if_empty,
    remove_newlines_and_tabs,
    validate,
)


HTTP_TOO_MANY_REQUESTS_ERROR = 429
HTTP_SERVICE_UNAVAILABLE_ERROR = 503
HTTP_GATEWAY_TIMEOUT_ERROR = 504
HTTP_INTERNAL_SERVER_ERROR = 500
HTTP_BAD_GATEWAY_ERROR = 502

T = TypeVar("T")


class EntraIDCredentials(BaseModel):
    """Entra ID credentials.

    Three key values are held in the Entra ID connector:
        - tenant_id: The tenant ID of the organization.
        - client_id: The client ID of the application.
        - client_secret: The client secret of the application.
        - scope: The scope of the application.

    Args:
        BaseModel (pydantic.BaseModel): Base class for data validation.

    """

    tenant_id: str
    client_id: str
    client_secret: str
    scope: str = "https://graph.microsoft.com/.default"


class EntraID(Source):
    """Entra ID (Azure AD) connector.

    It is responsible for fetching user-to-group memberships from Microsoft Graph.
    It uses the client credentials flow to authenticate and fetch the data.
    It supports pagination and retry logic for failed requests.

    Exposes a to_df() method that returns user-to-group memberships from
        Microsoft Graph.
    """

    def __init__(
        self,
        credentials: EntraIDCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Initialize the Entra ID source.

        Args:
            credentials (EntraIDCredentials | None): Explicit credentials for Entra ID.
                When None, credentials are loaded from the viadot config
                using `config_key`. Defaults to None.
            config_key (str | None): The key in the viadot config holding Entra ID
                credentials. Used only when `credentials` is None. Defaults to None.
            *args: Additional positional arguments forwarded to `Source`.
            **kwargs: Additional keyword arguments forwarded to `Source`.
        """
        credentials = credentials or get_source_credentials(config_key)
        validated_creds = dict(EntraIDCredentials(**credentials))
        super().__init__(*args, credentials=validated_creds, **kwargs)

    async def _get_access_token(self, session: aiohttp.ClientSession) -> str:
        """Obtain an access token from Microsoft Identity Platform for authentication.

        Args:
            session (aiohttp.ClientSession): The aiohttp session for HTTP requests.

        Returns:
            str: Bearer token for Microsoft Graph API access.

        Raises:
            aiohttp.ClientResponseError: If the token endpoint responds with
              a non-2xx status.
            aiohttp.ClientError: If a network-related error occurs.
        """
        token_url = f'https://login.microsoftonline.com/{self.credentials.get("tenant_id")}/oauth2/v2.0/token'

        token_data = {
            "grant_type": "client_credentials",
            "client_id": self.credentials.get("client_id"),
            "client_secret": self.credentials.get("client_secret"),
            "scope": self.credentials.get("scope"),
        }
        async with session.post(token_url, data=token_data) as resp:
            resp.raise_for_status()
            data = await resp.json()
            return data.get("access_token")

    async def _authorize_session(self, session: aiohttp.ClientSession) -> None:
        """Authorize the provided session by attaching a Bearer token to headers.

        Args:
            session (aiohttp.ClientSession): The aiohttp session to authorize.

        Raises:
            aiohttp.ClientResponseError:
                If obtaining an access token fails with non-2xx status.
            aiohttp.ClientError:
                If a network-related error occurs during token retrieval.
        """
        token = await self._get_access_token(session)
        session.headers.update({"Authorization": f"Bearer {token}"})

    async def _fetch_all_pages(
        self,
        session: aiohttp.ClientSession,
        url: str,
        max_retries: int = 5,
        backoff_factor: float = 0.5,
    ) -> list[dict]:
        """Fetch all results from paginated Microsoft Graph API endpoint async.

        Uses the '@odata.nextLink' property in the JSON response, which provides the URL
        for the next page of data. Iterates until no further pages remain.

        Args:
            session (aiohttp.ClientSession): The aiohttp session for HTTP requests.
            url (str): Initial Microsoft Graph API endpoint URL.
            max_retries (int): Maximum number of retries for 429/5xx responses.
            backoff_factor (float): Base delay for exponential backoff between retries.

        Returns:
            list[dict]: Aggr list of JSON objects from all pages of the API response.

        Raises:
            aiohttp.ClientResponseError: If the endpoint keeps failing after retries.
            aiohttp.ClientError: If a network-related error occurs.
        """
        items: list[dict] = []
        while url:
            attempt = 0
            while True:
                async with session.get(url) as resp:
                    # Retry on throttling and server errors
                    if (
                        resp.status
                        in (
                            HTTP_TOO_MANY_REQUESTS_ERROR,
                            HTTP_SERVICE_UNAVAILABLE_ERROR,
                            HTTP_GATEWAY_TIMEOUT_ERROR,
                        )
                        or HTTP_INTERNAL_SERVER_ERROR
                        <= resp.status
                        < HTTP_BAD_GATEWAY_ERROR
                    ):
                        attempt += 1
                        if attempt > max_retries:
                            resp.raise_for_status()
                        retry_after_header = resp.headers.get("Retry-After")
                        if retry_after_header:
                            try:
                                delay = float(retry_after_header)
                            except ValueError:
                                delay = backoff_factor * (2 ** (attempt - 1))
                        else:
                            delay = backoff_factor * (2 ** (attempt - 1))
                        self.logger.warning(
                            f"GET {url} -> HTTP {resp.status}. Retrying in {delay:.2f}s (attempt {attempt}/{max_retries})."
                        )
                        await asyncio.sleep(delay)
                        continue

                    # Raise immediately on other client errors
                    resp.raise_for_status()
                    data = await resp.json()
                    break

            chunk = data.get("value", [])
            items.extend(chunk)
            url = data.get("@odata.nextLink")

        return items

    async def _get_all_users(self, session: aiohttp.ClientSession) -> list:
        """Retrieve all users from Microsoft Graph API asynchronously.

        Args:
            session (aiohttp.ClientSession): The aiohttp session for HTTP requests.

        Returns:
            list: List of user JSON objects with at least id and mail.

        Raises:
            aiohttp.ClientResponseError: If listing users fails after retries.
            aiohttp.ClientError: If a network-related error occurs.
        """
        users_url = "https://graph.microsoft.com/v1.0/users?$select=id,mail"

        return await self._fetch_all_pages(session=session, url=users_url)

    async def _get_user_group_membership(
        self, session: aiohttp.ClientSession, user: dict
    ) -> list[dict]:
        """Retrieve all group memberships for a single user asynchronously.

        Args:
            session (aiohttp.ClientSession): The aiohttp session for HTTP requests.
            user (dict): User JSON object with at least 'id' and 'mail'.

        Returns:
            list[dict]: List of dicts with user info and group membership details.

        Raises:
            aiohttp.ClientResponseError: If the membership endpoint fails after retries.
            aiohttp.ClientError: If a network-related error occurs.
        """
        user_id = user.get("id", "")
        user_email = user.get("mail", "")
        groups_url = f"https://graph.microsoft.com/v1.0/users/{user_id}/memberOf"
        groups = await self._fetch_all_pages(session, groups_url)
        results = []
        for group in groups:
            results.append(
                {
                    "user_email": user_email,
                    "group_name": group.get("displayName", ""),
                }
            )
        return results

    async def _concat_user_groups_df(
        self,
        session: aiohttp.ClientSession,
        users: pd.DataFrame,
        output_fields: list,
        max_concurrent_requests: int = 50,
    ) -> pd.DataFrame:
        """Build a pandas DataFrame containing user-group membership.

        Args:
            session (aiohttp.ClientSession): The aiohttp session for HTTP requests.
            users (pd.DataFrame): DataFrame containing user records.
            output_fields (list): List of fields (columns) to include
                in the resulting DataFrame.
            max_concurrent_requests (int): Maximum number of concurrent HTTP requests.
                Increased to 50 for faster execution.

        Returns:
            pd.DataFrame: DataFrame with one row per user-group membership record.

        Raises:
            None: Errors during per-user fetching are logged and skipped.
        """
        all_rows: list[dict[str, Any]] = []
        semaphore = asyncio.Semaphore(max_concurrent_requests)

        async def sem_get_user_groups(user: dict) -> list:
            async with semaphore:
                try:
                    return await self._get_user_group_membership(session, user)
                except Exception:
                    self.logger.exception(
                        "Failed to get groups for user %s",
                        user.get("displayName", ""),
                    )
                    return []

        # Convert DataFrame rows to dicts to reduce overhead during async processing
        user_records = (
            users[["id", "mail"]].to_dict(orient="records") if not users.empty else []
        )
        tasks = [sem_get_user_groups(user) for user in user_records]
        for task in asyncio.as_completed(tasks):
            user_groups = await task
            if user_groups:
                all_rows.extend(user_groups)

        return pd.DataFrame(all_rows, columns=output_fields)

    async def _build_users_groups_df(
        self, max_concurrent_requests: int = 50
    ) -> pd.DataFrame:
        """Build a user-group membership DataFrame by querying users and groups.

        Args:
            max_concurrent_requests (int): Maximum number of concurrent HTTP requests
                to Microsoft Graph.

        Returns:
            pd.DataFrame: A DataFrame with columns `user_email` and `group_name`.

        Raises:
            aiohttp.ClientResponseError: If any Graph request fails after retries.
            aiohttp.ClientError: If a network-related error occurs.
        """
        timeout = aiohttp.ClientTimeout(total=120)
        connector = aiohttp.TCPConnector(limit=max_concurrent_requests)
        async with aiohttp.ClientSession(
            timeout=timeout, connector=connector
        ) as session:
            await self._authorize_session(session)
            users_list = await self._get_all_users(session)
            users_df = pd.DataFrame(users_list)
            output_fields = ["user_email", "group_name"]
            return await self._concat_user_groups_df(
                session=session,
                users=users_df,
                output_fields=output_fields,
                max_concurrent_requests=max_concurrent_requests,
            )

    def _run(self, coro: Coroutine[Any, Any, T]) -> T:
        """Safe runner to execute async coroutines from sync contexts.

        Args:
            coro (Coroutine): The coroutine to execute.

        Returns:
            Any: The result of the coroutine.

        Raises:
            RuntimeError: If the coroutine execution fails for reasons other than an
                already running event loop. When an event loop is running, a dedicated
                temporary loop is used to run the coroutine.
        """
        try:
            return asyncio.run(coro)
        except RuntimeError as e:
            # Fallback when already inside a running event loop
            if "asyncio.run() cannot be called from a running event loop" in str(e):
                loop = asyncio.new_event_loop()
                try:
                    asyncio.set_event_loop(loop)
                    return loop.run_until_complete(coro)
                finally:
                    try:
                        loop.close()
                    finally:
                        asyncio.set_event_loop(None)
            raise

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        tests: dict[str, Any] | None = None,
        max_concurrent_requests: int = 50,
    ) -> pd.DataFrame:
        """Produce a pandas DataFrame containing one row per user-group membership.

        Args:
            if_empty (Literal["warn", "skip", "fail"]): Policy for handling results.
            tests (dict[str, Any] | None): Optional tests to validate the resulting
                DataFrame. Passed through to `viadot.utils.validate`.
            max_concurrent_requests (int): Maximum number of concurrent HTTP requests
                 to Microsoft Graph.

        Returns:
            pd.DataFrame: Cleaned DataFrame with user-group memberships
                and viadot metadata columns.

        Raises:
            ValueError: If `if_empty="fail"` and the resulting DataFrame is empty.
            aiohttp.ClientResponseError: If any Graph request fails after retries.
            aiohttp.ClientError: If a network-related error occurs.
            Exception: Any exception raised by provided `tests` validation.
        """
        df = self._run(
            self._build_users_groups_df(max_concurrent_requests=max_concurrent_requests)
        )

        if df.empty:
            handle_if_empty(
                if_empty=if_empty,
                message="Resulting DataFrame is empty.",
                logger=self.logger,
            )

        df_clean = remove_newlines_and_tabs(df)

        if tests:
            validate(df=df_clean, tests=tests)

        return df_clean



"""'test_eurostat.py'."""

from contextlib import nullcontext as does_not_raise
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from viadot.sources import Eurostat


@pytest.fixture
def eurostat_instance():
    return Eurostat(dataset_code="TEIBS020", params=None, columns=None, tests=None)


@patch.object(Eurostat, "get_parameters_codes")
def test_validate_params_valid(mock_get_params, eurostat_instance):
    mock_get_params.return_value = {"unit": ["EUR", "USD"]}

    valid_params = {"unit": "EUR"}
    url = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/TEIBS020"

    with does_not_raise():
        eurostat_instance.validate_params("TEIBS020", url, valid_params)


@patch.object(Eurostat, "get_parameters_codes")
def test_validate_params_invalid_key(mock_get_params, eurostat_instance):
    mock_get_params.return_value = {"unit": ["EUR", "USD"]}

    invalid_params = {"invalid_key": "EUR"}
    url = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/TEIBS020"

    with pytest.raises(ValueError, match="Wrong parameters or codes were provided!"):
        eurostat_instance.validate_params("TEIBS020", url, invalid_params)


def test_eurostat_dictionary_to_df(eurostat_instance):
    mock_data = {
        "dimension": {
            "geo": {
                "category": {
                    "index": {"DE": 0, "FR": 1, "IT": 2},
                    "label": {"DE": "Germany", "FR": "France", "IT": "Italy"},
                }
            },
            "time": {
                "category": {
                    "index": {"2020": 0, "2021": 1, "2022": 2},
                    "label": {"2020": "2020", "2021": "2021", "2022": "2022"},
                }
            },
        },
        "value": {
            "0": 100,
            "1": 150,
            "2": 200,
            "3": 110,
            "4": 160,
            "5": 210,
            "6": 120,
            "7": 170,
            "8": 220,
        },
    }

    signals = [["geo", "time"]]

    # Call the method with signals as a list
    df = eurostat_instance.eurostat_dictionary_to_df(*signals, mock_data)

    assert isinstance(df, pd.DataFrame)
    assert df.shape == (9, 3)
    assert set(df.columns) == {"indicator", "geo", "time"}


@patch("viadot.utils.handle_api_request")
def test_get_parameters_codes_empty_response(
    mock_handle_api_request,
    eurostat_instance,
):
    mock_handle_api_request_obj = MagicMock()
    mock_handle_api_request_obj.json.return_value = {"id": [], "dimension": []}
    mock_handle_api_request.return_value = mock_handle_api_request_obj

    url = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/TEIBS020"

    result = eurostat_instance.get_parameters_codes(url)
    assert result == {}


@patch("viadot.utils.handle_api_request")
def test_get_parameters_codes_valid(
    mock_handle_api_request,
    eurostat_instance,
):
    mock_response = {
        "id": ["unit", "geo"],
        "dimension": {
            "unit": {
                "category": {
                    "index": {"EUR": 0, "USD": 1},
                    "label": {"Euro": 0, "US Dollar": 1},
                }
            },
            "geo": {
                "category": {
                    "index": {"FR": 0, "DE": 1},
                    "label": {"France": 0, "Germany": 1},
                }
            },
        },
    }

    mock_handle_api_request_obj = MagicMock()
    mock_handle_api_request_obj.json.return_value = mock_response
    mock_handle_api_request.return_value = mock_handle_api_request_obj

    url = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/TEIBS020"
    result = eurostat_instance.get_parameters_codes(url)

    expected_result = {"unit": ["EUR", "USD"], "geo": ["FR", "DE"]}

    assert result == expected_result


def test_to_df_invalid_columns_type(eurostat_instance):
    eurostat_instance.columns = "invalid_columns"  # Not a list

    with pytest.raises(
        TypeError, match="Requested columns should be provided as list of strings."
    ):
        eurostat_instance.to_df()


def test_to_df_invalid_params_type(eurostat_instance):
    eurostat_instance.params = "invalid_params"  # Not a dictionary

    with pytest.raises(TypeError, match="Params should be a dictionary."):
        eurostat_instance.to_df()



"""Amazon Redshift Spectrum connector."""

from typing import Literal

import pandas as pd


try:
    import awswrangler as wr
    import boto3
    import redshift_connector
except ModuleNotFoundError as e:
    msg = "Missing required modules to use RedshiftSpectrum source."
    raise ImportError(msg) from e

from typing import Any

from pydantic import BaseModel, root_validator

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.sources.base import Source


class RedshiftSpectrumCredentials(BaseModel):
    region_name: str  # The name of the AWS region.
    aws_access_key_id: str  # The AWS access key ID.
    aws_secret_access_key: str  # The AWS secret access key.
    profile_name: str | None = None  # The name of the IAM profile to use.
    endpoint_url: str | None = None  # The endpoint of the S3 service.

    # Below credentials are required only by some methods.
    #
    # The name of a AWS Secret holding credentials to a Redshift instance.
    # This AWS Secret has to include the following configuration:
    # host: Optional[str]
    # port: Optional[str] = "5439"
    # username: Optional[str]
    # password: Optional[str]
    # engine: Optional[str] = "redshift"
    # dbname: Optional[str]
    credentials_secret: str | None = None
    iam_role: str | None = None  # The IAM role to assume. Used by `create_schema()`.

    @root_validator(pre=True)
    def is_configured(cls, credentials: dict[str, Any]) -> dict[str, Any]:  # noqa: N805
        """Validate the credentials configuration."""
        profile_name = credentials.get("profile_name")
        region_name = credentials.get("region_name")
        aws_access_key_id = credentials.get("aws_access_key_id")
        aws_secret_access_key = credentials.get("aws_secret_access_key")

        profile_credential = profile_name and region_name
        direct_credential = aws_access_key_id and aws_secret_access_key and region_name

        if not (profile_credential or direct_credential):
            msg = "Either `profile_name` and `region_name`, or `aws_access_key_id`,"
            msg += " `aws_secret_access_key`, and `region_name` must be specified."
            raise CredentialError(msg)
        return credentials


class RedshiftSpectrum(Source):
    def __init__(
        self,
        credentials: RedshiftSpectrumCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """A class for working with Amazon Redshift Spectrum.

        Note that internally, AWS SDK refers to schemas as "databases", as external
        schemas correspond to AWS Glue databases. However, to keep consistent naming
        with all other viadot sources, we use the word "schema" instead.

        Args:
        credentials (RedshiftSpectrumCredentials, optional): RedshiftSpectrumCredentials
            credentials. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.

        Examples:
        ```python
        from viadot.sources import RedshiftSpectrum

        with RedshiftSpectrum(config_key="redshift_spectrum") as redshift:
            redshift.get_schemas()
        ```
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(RedshiftSpectrumCredentials(**raw_creds))

        super().__init__(*args, credentials=validated_creds, **kwargs)

        if not self.credentials:
            self.logger.debug(
                "Credentials not specified. Falling back to `boto3` default credentials."
            )

        if self.credentials:
            endpoint_url = self.credentials.get("endpoint_url")
            if endpoint_url:
                wr.config.s3_endpoint_url = endpoint_url

        self._session = None
        self._con = None

    def __enter__(self):  # noqa: D105
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # noqa: D105, ANN001
        if self._con:
            self._con.close()
            self._con = None

    @property
    def session(self) -> boto3.session.Session:
        """A singleton-like property for initiating an AWS session with boto3.

        Note that this is not an actual session, so it does not need to be closed.
        """
        if not self._session:
            self._session = boto3.session.Session(
                region_name=self.credentials.get("region_name"),
                profile_name=self.credentials.get("profile_name"),
                aws_access_key_id=self.credentials.get("aws_access_key_id"),
                aws_secret_access_key=self.credentials.get("aws_secret_access_key"),
            )
        return self._session

    @property
    def con(self) -> redshift_connector.Connection:
        """A singleton-like property for establishing a connection."""
        if not self._con:
            if self.credentials.get("credentials_secret"):
                self._con = wr.redshift.connect(
                    boto3_session=self.session,
                    timeout=10,
                    secret_id=self.credentials.get("credentials_secret"),
                )
            else:
                msg = "The `credentials_secret` config is required to connect to Redshift."
                raise ValueError(msg)
        return self._con

    def from_df(
        self,
        df: pd.DataFrame,
        to_path: str,
        schema: str,
        table: str,
        extension: Literal[".parquet", ".csv"] = ".parquet",
        if_exists: Literal["overwrite", "overwrite_partitions", "append"] = "overwrite",
        partition_cols: list[str] | None = None,
        sep: str = ",",
        **kwargs,
    ) -> None:
        """Upload a pandas `DataFrame` into a CSV or Parquet file.

        For a full list of available parameters, please refer to the official
        documentation:
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.to_parquet.html
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.to_csv.html

        Args:
            df (pd.DataFrame): Pandas `DataFrame`.
            to_path (str): Path to Amazon S3 folder where the table will be located. If
                needed, a bottom-level directory named f"{table}" is automatically
                created, so that files are always located in a folder named the same as
                the table.
            schema (str): The name of the schema.
            table (str): The name of the table to load the data into.
            extension (Literal[".parquet", ".csv"], optional): Required file type.
                Defaults to '.parquet'.
            if_exists (Literal["overwrite", "overwrite_partitions", "append"], optional
                ): 'overwrite' to recreate the table, 'overwrite_partitions' to only
                recreate the partitions, 'append' to append the data. Defaults to
                'overwrite'.
            partition_cols (List[str], optional): List of column names that will be used
                to create partitions. Only takes effect if dataset=True. Defaults to
                None.
            sep (str, optional): Field delimiter for the output file. Defaults to ','.
        """
        # Ensure files are in a directory named {table}.
        if not to_path.rstrip("/").endswith(table):
            to_path = to_path.rstrip("/") + "/" + table

        if extension == ".parquet":
            wr.s3.to_parquet(
                boto3_session=self.session,
                df=df,
                path=to_path,
                mode=if_exists,
                dataset=True,
                database=schema,
                table=table,
                partition_cols=partition_cols,
                **kwargs,
            )
        elif extension == ".csv":
            wr.s3.to_csv(
                boto3_session=self.session,
                df=df,
                path=to_path,
                dataset=True,
                database=schema,
                table=table,
                sep=sep,
                **kwargs,
            )
        else:
            msg = "Only CSV and parquet formats are supported."
            raise ValueError(msg)

    def to_df(
        self,
        schema: str,
        table: str,
        **kwargs,
    ) -> pd.DataFrame:
        """Read a table from an external schema into a pandas `DataFrame`.

        For a full list of available parameters, please refer to the official
        documentation:
        https://aws-sdk-pandas.readthedocs.io/en/3.0.0/stubs/awswrangler.s3.read_parquet_table.html

        Args:
            schema (str): The name of the schema.
            table (str): The name of the table to load.
        """
        return wr.s3.read_parquet_table(
            boto3_session=self.session,
            database=schema,
            table=table,
            **kwargs,
        )

    def drop_table(
        self,
        schema: str,
        table: str,
        remove_files: bool = True,
    ) -> None:
        """Drop a table from a specified external schema.

        Drops a table, including related files from Amazon S3, if specified.

        Args:
            schema (str): The name of the schema.
            table (str): The name of the table to drop.
            remove_files (bool, optional): If True, Amazon S3 file related to the table
                will be removed. Defaults to True.
        """
        if remove_files:
            table_location = wr.catalog.get_table_location(
                boto3_session=self.session,
                database=schema,
                table=table,
            )
            wr.s3.delete_objects(boto3_session=self.session, path=table_location)

        wr.catalog.delete_table_if_exists(
            boto3_session=self.session, database=schema, table=table
        )

    def get_tables(
        self,
        schema: str,
    ) -> list[str]:
        """Returns a list of tables in a specified schema.

        Args:
            schema (str): The name of the schema.
        """
        get_tables_query = f"SELECT t.tablename FROM SVV_EXTERNAL_TABLES t WHERE t.schemaname = '{schema}'"  # noqa: S608
        with self.con.cursor() as cursor:
            tables_info = cursor.execute(get_tables_query).fetchall()
        return [table_info[0] for table_info in tables_info]

    def _check_if_table_exists(self, schema: str, table: str) -> bool:
        """Check if a table exists in a specified Redshift Spectrum external schema.

        Args:
            schema (str): The name of the schema.
            table (str): The name of the table to verify.

        Returns:
            bool: Whether the table exists.
        """
        return table in self.get_tables(schema=schema)

    def create_schema(
        self,
        schema: str,
    ) -> None:
        """Create an external schema in Amazon Redshift Spectrum.

        This involves two steps:
        - creating a Glue database
        - creating an external schema in Redshift, pointing to above Glue database

        Args:
            schema (str): The name of the schema.
        """
        self._create_glue_database(database=schema, exist_ok=True)
        create_schema_query = f"""
create external schema if not exists "{schema}" from data catalog
database '{schema}'
iam_role '{self.credentials.get("iam_role")}'
region '{self.credentials.get("region_name")}'
CREATE EXTERNAL DATABASE IF NOT EXISTS;
"""
        with self.con.cursor() as cursor:
            cursor.execute(create_schema_query)
            self.con.commit()

    def _create_glue_database(
        self,
        database: str,
        exist_ok: bool = False,
    ):
        """Create an AWS Glue database.

        Args:
            database (str): The name of the database.
            exist_ok (bool, optional): Whether to skip if the database already exists.
                If set to False, will throw `AlreadyExistsException`. Defaults to False.
        """
        wr.catalog.create_database(
            name=database,
            boto3_session=self.session,
            exist_ok=exist_ok,
        )

    def get_schemas(self) -> list[str]:
        """Returns a list of schemas in the current Redshift Spectrum database."""
        # External Redshift schemas
        get_schemas_query = "SELECT schemaname FROM SVV_EXTERNAL_SCHEMAS"
        with self.con.cursor() as cursor:
            schema_names: tuple[list] = cursor.execute(get_schemas_query).fetchall()
        external_schemas = [schema_name[0] for schema_name in schema_names]

        # Glue databases.
        schema_infos = wr.catalog.get_databases(boto3_session=self.session)
        glue_schemas = [schema_info["Name"] for schema_info in schema_infos]

        # An external Redshift schema is a Spectrum schema only if it's also a Glue
        # database.
        return [schema for schema in external_schemas if schema in glue_schemas]

    def _check_if_schema_exists(self, schema: str) -> bool:
        """Check if a schema exists in Amazon Redshift Spectrum.

        Args:
            schema (str): The name of the schema.

        Returns:
            bool: Whether the schema exists.
        """
        return schema in self.get_schemas()

    def _is_spectrum_schema(self, schema: str) -> bool:
        """Check if a Redshift schema is a Spectrum schema.

        Args:
            schema (str): The name of the schema.

        Returns:
            bool: Whether the schema is a Spectrum schema.
        """
        return self._check_if_schema_exists(schema)

    def drop_schema(self, schema: str, drop_glue_database: bool = False) -> None:
        """Drop a Spectrum schema. If specified, also drop the underlying Glue database.

        Args:
            schema (str): The name of the schema.
        """
        if not self._is_spectrum_schema(schema):
            msg = f"Schema {schema} is not a Spectrum schema."
            raise ValueError(msg)

        drop_external_schema_query = f"DROP SCHEMA IF EXISTS {schema}"
        with self.con.cursor() as cursor:
            cursor.execute(drop_external_schema_query)
            self.con.commit()

        if drop_glue_database:
            wr.catalog.delete_database(
                name=schema,
                boto3_session=self.session,
            )



"""Task for running BCP shell command."""

import subprocess
from typing import Literal

from prefect import task
from prefect.logging import get_run_logger

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials


@task
def bcp(
    path: str,
    schema: str,
    table: str,
    chunksize: int = 5000,
    error_log_file_path: str = "./log_file.log",
    on_error: Literal["skip", "fail"] = "skip",
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> None:
    """Upload data from a CSV file into an SQLServer table using BCP.

    For more information on bcp (bulk copy program), see
    https://learn.microsoft.com/en-us/sql/tools/bcp-utility.

    Args:
        path (str):  Where to store the CSV data dump used for bulk upload to
        a database.
        schema (str): Destination schema. Defaults to None.
        table (str): Destination table. Defaults to None.
        chunksize (int, optional): Size of a chunk to use in the bcp function.
            Defaults to 5000.
        error_log_file_path (string, optional): Full path of an error file. Defaults
            to "./log_file.log".
        on_error (str, optional): What to do in case of a bcp error. Defaults to "skip".
        credentials_secret (str, optional): The name of the secret storing
            the credentials to the SQLServer. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials to the SQLServer. Defaults to None.

    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )
    fqn = f"{schema}.{table}" if schema else table
    server = credentials["server"]
    db_name = credentials["db_name"]
    uid = credentials["user"]
    pwd = credentials["password"]

    if "," in server:
        # A space after the comma is allowed in the ODBC connection string
        # but not in BCP's 'server' argument.
        server = server.replace(" ", "")

    if on_error == "skip":
        max_error = 0
    elif on_error == "fail":
        max_error = 1
    else:
        msg = "Please provide correct 'on_error' parameter value - 'skip' or 'fail'. "
        raise ValueError(msg)
    bcp_command = [
        "/opt/mssql-tools18/bin/bcp",
        fqn,
        "in",
        path,
        "-S",
        server,
        "-d",
        db_name,
        "-U",
        uid,
        "-P",
        pwd,
        "-u",
        "-b",
        str(chunksize),
        "-m",
        str(max_error),
        "-c",
        "-v",
        "-e",
        error_log_file_path,
        "-h",
        "TABLOCK",
        "-F",
        "2",
    ]

    result = subprocess.run(bcp_command, capture_output=True, text=True, check=False)  # noqa: S603

    logger = get_run_logger()
    if result.stdout:
        logger.info("BCP stdout:\n%s", result.stdout[:2000])
    if result.stderr:
        logger.warning("BCP stderr:\n%s", result.stderr[:2000])

    if result.returncode != 0:
        error_msg = (
            f"BCP failed with exit code {result.returncode}. "
            f"Stderr: {result.stderr[:500]}"
        )
        raise RuntimeError(error_msg)

    return result
